//go:build !ignore_autogenerated
// +build !ignore_autogenerated

/*
Copyright The Kubernetes Authors.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
*/

// Code generated by conversion-gen. DO NOT EDIT.

package v1beta1

import (
	unsafe "unsafe"

	corev1 "k8s.io/api/core/v1"
	resourcev1beta1 "k8s.io/api/resource/v1beta1"
	apiresource "k8s.io/apimachinery/pkg/api/resource"
	v1 "k8s.io/apimachinery/pkg/apis/meta/v1"
	conversion "k8s.io/apimachinery/pkg/conversion"
	runtime "k8s.io/apimachinery/pkg/runtime"
	types "k8s.io/apimachinery/pkg/types"
	core "k8s.io/kubernetes/pkg/apis/core"
	resource "k8s.io/kubernetes/pkg/apis/resource"
)

func init() {
	localSchemeBuilder.Register(RegisterConversions)
}

// RegisterConversions adds conversion functions to the given scheme.
// Public to allow building arbitrary schemes.
func RegisterConversions(s *runtime.Scheme) error {
	if err := s.AddGeneratedConversionFunc((*resourcev1beta1.AllocatedDeviceStatus)(nil), (*resource.AllocatedDeviceStatus)(nil), func(a, b interface{}, scope conversion.Scope) error {
		return Convert_v1beta1_AllocatedDeviceStatus_To_resource_AllocatedDeviceStatus(a.(*resourcev1beta1.AllocatedDeviceStatus), b.(*resource.AllocatedDeviceStatus), scope)
	}); err != nil {
		return err
	}
	if err := s.AddGeneratedConversionFunc((*resource.AllocatedDeviceStatus)(nil), (*resourcev1beta1.AllocatedDeviceStatus)(nil), func(a, b interface{}, scope conversion.Scope) error {
		return Convert_resource_AllocatedDeviceStatus_To_v1beta1_AllocatedDeviceStatus(a.(*resource.AllocatedDeviceStatus), b.(*resourcev1beta1.AllocatedDeviceStatus), scope)
	}); err != nil {
		return err
	}
	if err := s.AddGeneratedConversionFunc((*resourcev1beta1.AllocationResult)(nil), (*resource.AllocationResult)(nil), func(a, b interface{}, scope conversion.Scope) error {
		return Convert_v1beta1_AllocationResult_To_resource_AllocationResult(a.(*resourcev1beta1.AllocationResult), b.(*resource.AllocationResult), scope)
	}); err != nil {
		return err
	}
	if err := s.AddGeneratedConversionFunc((*resource.AllocationResult)(nil), (*resourcev1beta1.AllocationResult)(nil), func(a, b interface{}, scope conversion.Scope) error {
		return Convert_resource_AllocationResult_To_v1beta1_AllocationResult(a.(*resource.AllocationResult), b.(*resourcev1beta1.AllocationResult), scope)
	}); err != nil {
		return err
	}
	if err := s.AddGeneratedConversionFunc((*resourcev1beta1.CELDeviceSelector)(nil), (*resource.CELDeviceSelector)(nil), func(a, b interface{}, scope conversion.Scope) error {
		return Convert_v1beta1_CELDeviceSelector_To_resource_CELDeviceSelector(a.(*resourcev1beta1.CELDeviceSelector), b.(*resource.CELDeviceSelector), scope)
	}); err != nil {
		return err
	}
	if err := s.AddGeneratedConversionFunc((*resource.CELDeviceSelector)(nil), (*resourcev1beta1.CELDeviceSelector)(nil), func(a, b interface{}, scope conversion.Scope) error {
		return Convert_resource_CELDeviceSelector_To_v1beta1_CELDeviceSelector(a.(*resource.CELDeviceSelector), b.(*resourcev1beta1.CELDeviceSelector), scope)
	}); err != nil {
		return err
	}
	if err := s.AddGeneratedConversionFunc((*resourcev1beta1.CapacityRequestPolicy)(nil), (*resource.CapacityRequestPolicy)(nil), func(a, b interface{}, scope conversion.Scope) error {
		return Convert_v1beta1_CapacityRequestPolicy_To_resource_CapacityRequestPolicy(a.(*resourcev1beta1.CapacityRequestPolicy), b.(*resource.CapacityRequestPolicy), scope)
	}); err != nil {
		return err
	}
	if err := s.AddGeneratedConversionFunc((*resource.CapacityRequestPolicy)(nil), (*resourcev1beta1.CapacityRequestPolicy)(nil), func(a, b interface{}, scope conversion.Scope) error {
		return Convert_resource_CapacityRequestPolicy_To_v1beta1_CapacityRequestPolicy(a.(*resource.CapacityRequestPolicy), b.(*resourcev1beta1.CapacityRequestPolicy), scope)
	}); err != nil {
		return err
	}
	if err := s.AddGeneratedConversionFunc((*resourcev1beta1.CapacityRequestPolicyRange)(nil), (*resource.CapacityRequestPolicyRange)(nil), func(a, b interface{}, scope conversion.Scope) error {
		return Convert_v1beta1_CapacityRequestPolicyRange_To_resource_CapacityRequestPolicyRange(a.(*resourcev1beta1.CapacityRequestPolicyRange), b.(*resource.CapacityRequestPolicyRange), scope)
	}); err != nil {
		return err
	}
	if err := s.AddGeneratedConversionFunc((*resource.CapacityRequestPolicyRange)(nil), (*resourcev1beta1.CapacityRequestPolicyRange)(nil), func(a, b interface{}, scope conversion.Scope) error {
		return Convert_resource_CapacityRequestPolicyRange_To_v1beta1_CapacityRequestPolicyRange(a.(*resource.CapacityRequestPolicyRange), b.(*resourcev1beta1.CapacityRequestPolicyRange), scope)
	}); err != nil {
		return err
	}
	if err := s.AddGeneratedConversionFunc((*resourcev1beta1.CapacityRequirements)(nil), (*resource.CapacityRequirements)(nil), func(a, b interface{}, scope conversion.Scope) error {
		return Convert_v1beta1_CapacityRequirements_To_resource_CapacityRequirements(a.(*resourcev1beta1.CapacityRequirements), b.(*resource.CapacityRequirements), scope)
	}); err != nil {
		return err
	}
	if err := s.AddGeneratedConversionFunc((*resource.CapacityRequirements)(nil), (*resourcev1beta1.CapacityRequirements)(nil), func(a, b interface{}, scope conversion.Scope) error {
		return Convert_resource_CapacityRequirements_To_v1beta1_CapacityRequirements(a.(*resource.CapacityRequirements), b.(*resourcev1beta1.CapacityRequirements), scope)
	}); err != nil {
		return err
	}
	if err := s.AddGeneratedConversionFunc((*resourcev1beta1.Counter)(nil), (*resource.Counter)(nil), func(a, b interface{}, scope conversion.Scope) error {
		return Convert_v1beta1_Counter_To_resource_Counter(a.(*resourcev1beta1.Counter), b.(*resource.Counter), scope)
	}); err != nil {
		return err
	}
	if err := s.AddGeneratedConversionFunc((*resource.Counter)(nil), (*resourcev1beta1.Counter)(nil), func(a, b interface{}, scope conversion.Scope) error {
		return Convert_resource_Counter_To_v1beta1_Counter(a.(*resource.Counter), b.(*resourcev1beta1.Counter), scope)
	}); err != nil {
		return err
	}
	if err := s.AddGeneratedConversionFunc((*resourcev1beta1.CounterSet)(nil), (*resource.CounterSet)(nil), func(a, b interface{}, scope conversion.Scope) error {
		return Convert_v1beta1_CounterSet_To_resource_CounterSet(a.(*resourcev1beta1.CounterSet), b.(*resource.CounterSet), scope)
	}); err != nil {
		return err
	}
	if err := s.AddGeneratedConversionFunc((*resource.CounterSet)(nil), (*resourcev1beta1.CounterSet)(nil), func(a, b interface{}, scope conversion.Scope) error {
		return Convert_resource_CounterSet_To_v1beta1_CounterSet(a.(*resource.CounterSet), b.(*resourcev1beta1.CounterSet), scope)
	}); err != nil {
		return err
	}
	if err := s.AddGeneratedConversionFunc((*resourcev1beta1.DeviceAllocationConfiguration)(nil), (*resource.DeviceAllocationConfiguration)(nil), func(a, b interface{}, scope conversion.Scope) error {
		return Convert_v1beta1_DeviceAllocationConfiguration_To_resource_DeviceAllocationConfiguration(a.(*resourcev1beta1.DeviceAllocationConfiguration), b.(*resource.DeviceAllocationConfiguration), scope)
	}); err != nil {
		return err
	}
	if err := s.AddGeneratedConversionFunc((*resource.DeviceAllocationConfiguration)(nil), (*resourcev1beta1.DeviceAllocationConfiguration)(nil), func(a, b interface{}, scope conversion.Scope) error {
		return Convert_resource_DeviceAllocationConfiguration_To_v1beta1_DeviceAllocationConfiguration(a.(*resource.DeviceAllocationConfiguration), b.(*resourcev1beta1.DeviceAllocationConfiguration), scope)
	}); err != nil {
		return err
	}
	if err := s.AddGeneratedConversionFunc((*resourcev1beta1.DeviceAllocationResult)(nil), (*resource.DeviceAllocationResult)(nil), func(a, b interface{}, scope conversion.Scope) error {
		return Convert_v1beta1_DeviceAllocationResult_To_resource_DeviceAllocationResult(a.(*resourcev1beta1.DeviceAllocationResult), b.(*resource.DeviceAllocationResult), scope)
	}); err != nil {
		return err
	}
	if err := s.AddGeneratedConversionFunc((*resource.DeviceAllocationResult)(nil), (*resourcev1beta1.DeviceAllocationResult)(nil), func(a, b interface{}, scope conversion.Scope) error {
		return Convert_resource_DeviceAllocationResult_To_v1beta1_DeviceAllocationResult(a.(*resource.DeviceAllocationResult), b.(*resourcev1beta1.DeviceAllocationResult), scope)
	}); err != nil {
		return err
	}
	if err := s.AddGeneratedConversionFunc((*resourcev1beta1.DeviceAttribute)(nil), (*resource.DeviceAttribute)(nil), func(a, b interface{}, scope conversion.Scope) error {
		return Convert_v1beta1_DeviceAttribute_To_resource_DeviceAttribute(a.(*resourcev1beta1.DeviceAttribute), b.(*resource.DeviceAttribute), scope)
	}); err != nil {
		return err
	}
	if err := s.AddGeneratedConversionFunc((*resource.DeviceAttribute)(nil), (*resourcev1beta1.DeviceAttribute)(nil), func(a, b interface{}, scope conversion.Scope) error {
		return Convert_resource_DeviceAttribute_To_v1beta1_DeviceAttribute(a.(*resource.DeviceAttribute), b.(*resourcev1beta1.DeviceAttribute), scope)
	}); err != nil {
		return err
	}
	if err := s.AddGeneratedConversionFunc((*resourcev1beta1.DeviceCapacity)(nil), (*resource.DeviceCapacity)(nil), func(a, b interface{}, scope conversion.Scope) error {
		return Convert_v1beta1_DeviceCapacity_To_resource_DeviceCapacity(a.(*resourcev1beta1.DeviceCapacity), b.(*resource.DeviceCapacity), scope)
	}); err != nil {
		return err
	}
	if err := s.AddGeneratedConversionFunc((*resource.DeviceCapacity)(nil), (*resourcev1beta1.DeviceCapacity)(nil), func(a, b interface{}, scope conversion.Scope) error {
		return Convert_resource_DeviceCapacity_To_v1beta1_DeviceCapacity(a.(*resource.DeviceCapacity), b.(*resourcev1beta1.DeviceCapacity), scope)
	}); err != nil {
		return err
	}
	if err := s.AddGeneratedConversionFunc((*resourcev1beta1.DeviceClaim)(nil), (*resource.DeviceClaim)(nil), func(a, b interface{}, scope conversion.Scope) error {
		return Convert_v1beta1_DeviceClaim_To_resource_DeviceClaim(a.(*resourcev1beta1.DeviceClaim), b.(*resource.DeviceClaim), scope)
	}); err != nil {
		return err
	}
	if err := s.AddGeneratedConversionFunc((*resource.DeviceClaim)(nil), (*resourcev1beta1.DeviceClaim)(nil), func(a, b interface{}, scope conversion.Scope) error {
		return Convert_resource_DeviceClaim_To_v1beta1_DeviceClaim(a.(*resource.DeviceClaim), b.(*resourcev1beta1.DeviceClaim), scope)
	}); err != nil {
		return err
	}
	if err := s.AddGeneratedConversionFunc((*resourcev1beta1.DeviceClaimConfiguration)(nil), (*resource.DeviceClaimConfiguration)(nil), func(a, b interface{}, scope conversion.Scope) error {
		return Convert_v1beta1_DeviceClaimConfiguration_To_resource_DeviceClaimConfiguration(a.(*resourcev1beta1.DeviceClaimConfiguration), b.(*resource.DeviceClaimConfiguration), scope)
	}); err != nil {
		return err
	}
	if err := s.AddGeneratedConversionFunc((*resource.DeviceClaimConfiguration)(nil), (*resourcev1beta1.DeviceClaimConfiguration)(nil), func(a, b interface{}, scope conversion.Scope) error {
		return Convert_resource_DeviceClaimConfiguration_To_v1beta1_DeviceClaimConfiguration(a.(*resource.DeviceClaimConfiguration), b.(*resourcev1beta1.DeviceClaimConfiguration), scope)
	}); err != nil {
		return err
	}
	if err := s.AddGeneratedConversionFunc((*resourcev1beta1.DeviceClass)(nil), (*resource.DeviceClass)(nil), func(a, b interface{}, scope conversion.Scope) error {
		return Convert_v1beta1_DeviceClass_To_resource_DeviceClass(a.(*resourcev1beta1.DeviceClass), b.(*resource.DeviceClass), scope)
	}); err != nil {
		return err
	}
	if err := s.AddGeneratedConversionFunc((*resource.DeviceClass)(nil), (*resourcev1beta1.DeviceClass)(nil), func(a, b interface{}, scope conversion.Scope) error {
		return Convert_resource_DeviceClass_To_v1beta1_DeviceClass(a.(*resource.DeviceClass), b.(*resourcev1beta1.DeviceClass), scope)
	}); err != nil {
		return err
	}
	if err := s.AddGeneratedConversionFunc((*resourcev1beta1.DeviceClassConfiguration)(nil), (*resource.DeviceClassConfiguration)(nil), func(a, b interface{}, scope conversion.Scope) error {
		return Convert_v1beta1_DeviceClassConfiguration_To_resource_DeviceClassConfiguration(a.(*resourcev1beta1.DeviceClassConfiguration), b.(*resource.DeviceClassConfiguration), scope)
	}); err != nil {
		return err
	}
	if err := s.AddGeneratedConversionFunc((*resource.DeviceClassConfiguration)(nil), (*resourcev1beta1.DeviceClassConfiguration)(nil), func(a, b interface{}, scope conversion.Scope) error {
		return Convert_resource_DeviceClassConfiguration_To_v1beta1_DeviceClassConfiguration(a.(*resource.DeviceClassConfiguration), b.(*resourcev1beta1.DeviceClassConfiguration), scope)
	}); err != nil {
		return err
	}
	if err := s.AddGeneratedConversionFunc((*resourcev1beta1.DeviceClassList)(nil), (*resource.DeviceClassList)(nil), func(a, b interface{}, scope conversion.Scope) error {
		return Convert_v1beta1_DeviceClassList_To_resource_DeviceClassList(a.(*resourcev1beta1.DeviceClassList), b.(*resource.DeviceClassList), scope)
	}); err != nil {
		return err
	}
	if err := s.AddGeneratedConversionFunc((*resource.DeviceClassList)(nil), (*resourcev1beta1.DeviceClassList)(nil), func(a, b interface{}, scope conversion.Scope) error {
		return Convert_resource_DeviceClassList_To_v1beta1_DeviceClassList(a.(*resource.DeviceClassList), b.(*resourcev1beta1.DeviceClassList), scope)
	}); err != nil {
		return err
	}
	if err := s.AddGeneratedConversionFunc((*resourcev1beta1.DeviceClassSpec)(nil), (*resource.DeviceClassSpec)(nil), func(a, b interface{}, scope conversion.Scope) error {
		return Convert_v1beta1_DeviceClassSpec_To_resource_DeviceClassSpec(a.(*resourcev1beta1.DeviceClassSpec), b.(*resource.DeviceClassSpec), scope)
	}); err != nil {
		return err
	}
	if err := s.AddGeneratedConversionFunc((*resource.DeviceClassSpec)(nil), (*resourcev1beta1.DeviceClassSpec)(nil), func(a, b interface{}, scope conversion.Scope) error {
		return Convert_resource_DeviceClassSpec_To_v1beta1_DeviceClassSpec(a.(*resource.DeviceClassSpec), b.(*resourcev1beta1.DeviceClassSpec), scope)
	}); err != nil {
		return err
	}
	if err := s.AddGeneratedConversionFunc((*resourcev1beta1.DeviceConfiguration)(nil), (*resource.DeviceConfiguration)(nil), func(a, b interface{}, scope conversion.Scope) error {
		return Convert_v1beta1_DeviceConfiguration_To_resource_DeviceConfiguration(a.(*resourcev1beta1.DeviceConfiguration), b.(*resource.DeviceConfiguration), scope)
	}); err != nil {
		return err
	}
	if err := s.AddGeneratedConversionFunc((*resource.DeviceConfiguration)(nil), (*resourcev1beta1.DeviceConfiguration)(nil), func(a, b interface{}, scope conversion.Scope) error {
		return Convert_resource_DeviceConfiguration_To_v1beta1_DeviceConfiguration(a.(*resource.DeviceConfiguration), b.(*resourcev1beta1.DeviceConfiguration), scope)
	}); err != nil {
		return err
	}
	if err := s.AddGeneratedConversionFunc((*resourcev1beta1.DeviceConstraint)(nil), (*resource.DeviceConstraint)(nil), func(a, b interface{}, scope conversion.Scope) error {
		return Convert_v1beta1_DeviceConstraint_To_resource_DeviceConstraint(a.(*resourcev1beta1.DeviceConstraint), b.(*resource.DeviceConstraint), scope)
	}); err != nil {
		return err
	}
	if err := s.AddGeneratedConversionFunc((*resource.DeviceConstraint)(nil), (*resourcev1beta1.DeviceConstraint)(nil), func(a, b interface{}, scope conversion.Scope) error {
		return Convert_resource_DeviceConstraint_To_v1beta1_DeviceConstraint(a.(*resource.DeviceConstraint), b.(*resourcev1beta1.DeviceConstraint), scope)
	}); err != nil {
		return err
	}
	if err := s.AddGeneratedConversionFunc((*resourcev1beta1.DeviceCounterConsumption)(nil), (*resource.DeviceCounterConsumption)(nil), func(a, b interface{}, scope conversion.Scope) error {
		return Convert_v1beta1_DeviceCounterConsumption_To_resource_DeviceCounterConsumption(a.(*resourcev1beta1.DeviceCounterConsumption), b.(*resource.DeviceCounterConsumption), scope)
	}); err != nil {
		return err
	}
	if err := s.AddGeneratedConversionFunc((*resource.DeviceCounterConsumption)(nil), (*resourcev1beta1.DeviceCounterConsumption)(nil), func(a, b interface{}, scope conversion.Scope) error {
		return Convert_resource_DeviceCounterConsumption_To_v1beta1_DeviceCounterConsumption(a.(*resource.DeviceCounterConsumption), b.(*resourcev1beta1.DeviceCounterConsumption), scope)
	}); err != nil {
		return err
	}
	if err := s.AddGeneratedConversionFunc((*resourcev1beta1.DeviceRequestAllocationResult)(nil), (*resource.DeviceRequestAllocationResult)(nil), func(a, b interface{}, scope conversion.Scope) error {
		return Convert_v1beta1_DeviceRequestAllocationResult_To_resource_DeviceRequestAllocationResult(a.(*resourcev1beta1.DeviceRequestAllocationResult), b.(*resource.DeviceRequestAllocationResult), scope)
	}); err != nil {
		return err
	}
	if err := s.AddGeneratedConversionFunc((*resource.DeviceRequestAllocationResult)(nil), (*resourcev1beta1.DeviceRequestAllocationResult)(nil), func(a, b interface{}, scope conversion.Scope) error {
		return Convert_resource_DeviceRequestAllocationResult_To_v1beta1_DeviceRequestAllocationResult(a.(*resource.DeviceRequestAllocationResult), b.(*resourcev1beta1.DeviceRequestAllocationResult), scope)
	}); err != nil {
		return err
	}
	if err := s.AddGeneratedConversionFunc((*resourcev1beta1.DeviceSelector)(nil), (*resource.DeviceSelector)(nil), func(a, b interface{}, scope conversion.Scope) error {
		return Convert_v1beta1_DeviceSelector_To_resource_DeviceSelector(a.(*resourcev1beta1.DeviceSelector), b.(*resource.DeviceSelector), scope)
	}); err != nil {
		return err
	}
	if err := s.AddGeneratedConversionFunc((*resource.DeviceSelector)(nil), (*resourcev1beta1.DeviceSelector)(nil), func(a, b interface{}, scope conversion.Scope) error {
		return Convert_resource_DeviceSelector_To_v1beta1_DeviceSelector(a.(*resource.DeviceSelector), b.(*resourcev1beta1.DeviceSelector), scope)
	}); err != nil {
		return err
	}
	if err := s.AddGeneratedConversionFunc((*resourcev1beta1.DeviceSubRequest)(nil), (*resource.DeviceSubRequest)(nil), func(a, b interface{}, scope conversion.Scope) error {
		return Convert_v1beta1_DeviceSubRequest_To_resource_DeviceSubRequest(a.(*resourcev1beta1.DeviceSubRequest), b.(*resource.DeviceSubRequest), scope)
	}); err != nil {
		return err
	}
	if err := s.AddGeneratedConversionFunc((*resource.DeviceSubRequest)(nil), (*resourcev1beta1.DeviceSubRequest)(nil), func(a, b interface{}, scope conversion.Scope) error {
		return Convert_resource_DeviceSubRequest_To_v1beta1_DeviceSubRequest(a.(*resource.DeviceSubRequest), b.(*resourcev1beta1.DeviceSubRequest), scope)
	}); err != nil {
		return err
	}
	if err := s.AddGeneratedConversionFunc((*resourcev1beta1.DeviceTaint)(nil), (*resource.DeviceTaint)(nil), func(a, b interface{}, scope conversion.Scope) error {
		return Convert_v1beta1_DeviceTaint_To_resource_DeviceTaint(a.(*resourcev1beta1.DeviceTaint), b.(*resource.DeviceTaint), scope)
	}); err != nil {
		return err
	}
	if err := s.AddGeneratedConversionFunc((*resource.DeviceTaint)(nil), (*resourcev1beta1.DeviceTaint)(nil), func(a, b interface{}, scope conversion.Scope) error {
		return Convert_resource_DeviceTaint_To_v1beta1_DeviceTaint(a.(*resource.DeviceTaint), b.(*resourcev1beta1.DeviceTaint), scope)
	}); err != nil {
		return err
	}
	if err := s.AddGeneratedConversionFunc((*resourcev1beta1.DeviceToleration)(nil), (*resource.DeviceToleration)(nil), func(a, b interface{}, scope conversion.Scope) error {
		return Convert_v1beta1_DeviceToleration_To_resource_DeviceToleration(a.(*resourcev1beta1.DeviceToleration), b.(*resource.DeviceToleration), scope)
	}); err != nil {
		return err
	}
	if err := s.AddGeneratedConversionFunc((*resource.DeviceToleration)(nil), (*resourcev1beta1.DeviceToleration)(nil), func(a, b interface{}, scope conversion.Scope) error {
		return Convert_resource_DeviceToleration_To_v1beta1_DeviceToleration(a.(*resource.DeviceToleration), b.(*resourcev1beta1.DeviceToleration), scope)
	}); err != nil {
		return err
	}
	if err := s.AddGeneratedConversionFunc((*resourcev1beta1.NetworkDeviceData)(nil), (*resource.NetworkDeviceData)(nil), func(a, b interface{}, scope conversion.Scope) error {
		return Convert_v1beta1_NetworkDeviceData_To_resource_NetworkDeviceData(a.(*resourcev1beta1.NetworkDeviceData), b.(*resource.NetworkDeviceData), scope)
	}); err != nil {
		return err
	}
	if err := s.AddGeneratedConversionFunc((*resource.NetworkDeviceData)(nil), (*resourcev1beta1.NetworkDeviceData)(nil), func(a, b interface{}, scope conversion.Scope) error {
		return Convert_resource_NetworkDeviceData_To_v1beta1_NetworkDeviceData(a.(*resource.NetworkDeviceData), b.(*resourcev1beta1.NetworkDeviceData), scope)
	}); err != nil {
		return err
	}
	if err := s.AddGeneratedConversionFunc((*resourcev1beta1.NodeAllocatableResourceMapping)(nil), (*resource.NodeAllocatableResourceMapping)(nil), func(a, b interface{}, scope conversion.Scope) error {
		return Convert_v1beta1_NodeAllocatableResourceMapping_To_resource_NodeAllocatableResourceMapping(a.(*resourcev1beta1.NodeAllocatableResourceMapping), b.(*resource.NodeAllocatableResourceMapping), scope)
	}); err != nil {
		return err
	}
	if err := s.AddGeneratedConversionFunc((*resource.NodeAllocatableResourceMapping)(nil), (*resourcev1beta1.NodeAllocatableResourceMapping)(nil), func(a, b interface{}, scope conversion.Scope) error {
		return Convert_resource_NodeAllocatableResourceMapping_To_v1beta1_NodeAllocatableResourceMapping(a.(*resource.NodeAllocatableResourceMapping), b.(*resourcev1beta1.NodeAllocatableResourceMapping), scope)
	}); err != nil {
		return err
	}
	if err := s.AddGeneratedConversionFunc((*resourcev1beta1.OpaqueDeviceConfiguration)(nil), (*resource.OpaqueDeviceConfiguration)(nil), func(a, b interface{}, scope conversion.Scope) error {
		return Convert_v1beta1_OpaqueDeviceConfiguration_To_resource_OpaqueDeviceConfiguration(a.(*resourcev1beta1.OpaqueDeviceConfiguration), b.(*resource.OpaqueDeviceConfiguration), scope)
	}); err != nil {
		return err
	}
	if err := s.AddGeneratedConversionFunc((*resource.OpaqueDeviceConfiguration)(nil), (*resourcev1beta1.OpaqueDeviceConfiguration)(nil), func(a, b interface{}, scope conversion.Scope) error {
		return Convert_resource_OpaqueDeviceConfiguration_To_v1beta1_OpaqueDeviceConfiguration(a.(*resource.OpaqueDeviceConfiguration), b.(*resourcev1beta1.OpaqueDeviceConfiguration), scope)
	}); err != nil {
		return err
	}
	if err := s.AddGeneratedConversionFunc((*resourcev1beta1.ResourceClaim)(nil), (*resource.ResourceClaim)(nil), func(a, b interface{}, scope conversion.Scope) error {
		return Convert_v1beta1_ResourceClaim_To_resource_ResourceClaim(a.(*resourcev1beta1.ResourceClaim), b.(*resource.ResourceClaim), scope)
	}); err != nil {
		return err
	}
	if err := s.AddGeneratedConversionFunc((*resource.ResourceClaim)(nil), (*resourcev1beta1.ResourceClaim)(nil), func(a, b interface{}, scope conversion.Scope) error {
		return Convert_resource_ResourceClaim_To_v1beta1_ResourceClaim(a.(*resource.ResourceClaim), b.(*resourcev1beta1.ResourceClaim), scope)
	}); err != nil {
		return err
	}
	if err := s.AddGeneratedConversionFunc((*resourcev1beta1.ResourceClaimConsumerReference)(nil), (*resource.ResourceClaimConsumerReference)(nil), func(a, b interface{}, scope conversion.Scope) error {
		return Convert_v1beta1_ResourceClaimConsumerReference_To_resource_ResourceClaimConsumerReference(a.(*resourcev1beta1.ResourceClaimConsumerReference), b.(*resource.ResourceClaimConsumerReference), scope)
	}); err != nil {
		return err
	}
	if err := s.AddGeneratedConversionFunc((*resource.ResourceClaimConsumerReference)(nil), (*resourcev1beta1.ResourceClaimConsumerReference)(nil), func(a, b interface{}, scope conversion.Scope) error {
		return Convert_resource_ResourceClaimConsumerReference_To_v1beta1_ResourceClaimConsumerReference(a.(*resource.ResourceClaimConsumerReference), b.(*resourcev1beta1.ResourceClaimConsumerReference), scope)
	}); err != nil {
		return err
	}
	if err := s.AddGeneratedConversionFunc((*resourcev1beta1.ResourceClaimList)(nil), (*resource.ResourceClaimList)(nil), func(a, b interface{}, scope conversion.Scope) error {
		return Convert_v1beta1_ResourceClaimList_To_resource_ResourceClaimList(a.(*resourcev1beta1.ResourceClaimList), b.(*resource.ResourceClaimList), scope)
	}); err != nil {
		return err
	}
	if err := s.AddGeneratedConversionFunc((*resource.ResourceClaimList)(nil), (*resourcev1beta1.ResourceClaimList)(nil), func(a, b interface{}, scope conversion.Scope) error {
		return Convert_resource_ResourceClaimList_To_v1beta1_ResourceClaimList(a.(*resource.ResourceClaimList), b.(*resourcev1beta1.ResourceClaimList), scope)
	}); err != nil {
		return err
	}
	if err := s.AddGeneratedConversionFunc((*resourcev1beta1.ResourceClaimSpec)(nil), (*resource.ResourceClaimSpec)(nil), func(a, b interface{}, scope conversion.Scope) error {
		return Convert_v1beta1_ResourceClaimSpec_To_resource_ResourceClaimSpec(a.(*resourcev1beta1.ResourceClaimSpec), b.(*resource.ResourceClaimSpec), scope)
	}); err != nil {
		return err
	}
	if err := s.AddGeneratedConversionFunc((*resource.ResourceClaimSpec)(nil), (*resourcev1beta1.ResourceClaimSpec)(nil), func(a, b interface{}, scope conversion.Scope) error {
		return Convert_resource_ResourceClaimSpec_To_v1beta1_ResourceClaimSpec(a.(*resource.ResourceClaimSpec), b.(*resourcev1beta1.ResourceClaimSpec), scope)
	}); err != nil {
		return err
	}
	if err := s.AddGeneratedConversionFunc((*resourcev1beta1.ResourceClaimStatus)(nil), (*resource.ResourceClaimStatus)(nil), func(a, b interface{}, scope conversion.Scope) error {
		return Convert_v1beta1_ResourceClaimStatus_To_resource_ResourceClaimStatus(a.(*resourcev1beta1.ResourceClaimStatus), b.(*resource.ResourceClaimStatus), scope)
	}); err != nil {
		return err
	}
	if err := s.AddGeneratedConversionFunc((*resource.ResourceClaimStatus)(nil), (*resourcev1beta1.ResourceClaimStatus)(nil), func(a, b interface{}, scope conversion.Scope) error {
		return Convert_resource_ResourceClaimStatus_To_v1beta1_ResourceClaimStatus(a.(*resource.ResourceClaimStatus), b.(*resourcev1beta1.ResourceClaimStatus), scope)
	}); err != nil {
		return err
	}
	if err := s.AddGeneratedConversionFunc((*resourcev1beta1.ResourceClaimTemplate)(nil), (*resource.ResourceClaimTemplate)(nil), func(a, b interface{}, scope conversion.Scope) error {
		return Convert_v1beta1_ResourceClaimTemplate_To_resource_ResourceClaimTemplate(a.(*resourcev1beta1.ResourceClaimTemplate), b.(*resource.ResourceClaimTemplate), scope)
	}); err != nil {
		return err
	}
	if err := s.AddGeneratedConversionFunc((*resource.ResourceClaimTemplate)(nil), (*resourcev1beta1.ResourceClaimTemplate)(nil), func(a, b interface{}, scope conversion.Scope) error {
		return Convert_resource_ResourceClaimTemplate_To_v1beta1_ResourceClaimTemplate(a.(*resource.ResourceClaimTemplate), b.(*resourcev1beta1.ResourceClaimTemplate), scope)
	}); err != nil {
		return err
	}
	if err := s.AddGeneratedConversionFunc((*resourcev1beta1.ResourceClaimTemplateList)(nil), (*resource.ResourceClaimTemplateList)(nil), func(a, b interface{}, scope conversion.Scope) error {
		return Convert_v1beta1_ResourceClaimTemplateList_To_resource_ResourceClaimTemplateList(a.(*resourcev1beta1.ResourceClaimTemplateList), b.(*resource.ResourceClaimTemplateList), scope)
	}); err != nil {
		return err
	}
	if err := s.AddGeneratedConversionFunc((*resource.ResourceClaimTemplateList)(nil), (*resourcev1beta1.ResourceClaimTemplateList)(nil), func(a, b interface{}, scope conversion.Scope) error {
		return Convert_resource_ResourceClaimTemplateList_To_v1beta1_ResourceClaimTemplateList(a.(*resource.ResourceClaimTemplateList), b.(*resourcev1beta1.ResourceClaimTemplateList), scope)
	}); err != nil {
		return err
	}
	if err := s.AddGeneratedConversionFunc((*resourcev1beta1.ResourceClaimTemplateSpec)(nil), (*resource.ResourceClaimTemplateSpec)(nil), func(a, b interface{}, scope conversion.Scope) error {
		return Convert_v1beta1_ResourceClaimTemplateSpec_To_resource_ResourceClaimTemplateSpec(a.(*resourcev1beta1.ResourceClaimTemplateSpec), b.(*resource.ResourceClaimTemplateSpec), scope)
	}); err != nil {
		return err
	}
	if err := s.AddGeneratedConversionFunc((*resource.ResourceClaimTemplateSpec)(nil), (*resourcev1beta1.ResourceClaimTemplateSpec)(nil), func(a, b interface{}, scope conversion.Scope) error {
		return Convert_resource_ResourceClaimTemplateSpec_To_v1beta1_ResourceClaimTemplateSpec(a.(*resource.ResourceClaimTemplateSpec), b.(*resourcev1beta1.ResourceClaimTemplateSpec), scope)
	}); err != nil {
		return err
	}
	if err := s.AddGeneratedConversionFunc((*resourcev1beta1.ResourcePool)(nil), (*resource.ResourcePool)(nil), func(a, b interface{}, scope conversion.Scope) error {
		return Convert_v1beta1_ResourcePool_To_resource_ResourcePool(a.(*resourcev1beta1.ResourcePool), b.(*resource.ResourcePool), scope)
	}); err != nil {
		return err
	}
	if err := s.AddGeneratedConversionFunc((*resource.ResourcePool)(nil), (*resourcev1beta1.ResourcePool)(nil), func(a, b interface{}, scope conversion.Scope) error {
		return Convert_resource_ResourcePool_To_v1beta1_ResourcePool(a.(*resource.ResourcePool), b.(*resourcev1beta1.ResourcePool), scope)
	}); err != nil {
		return err
	}
	if err := s.AddGeneratedConversionFunc((*resourcev1beta1.ResourceSlice)(nil), (*resource.ResourceSlice)(nil), func(a, b interface{}, scope conversion.Scope) error {
		return Convert_v1beta1_ResourceSlice_To_resource_ResourceSlice(a.(*resourcev1beta1.ResourceSlice), b.(*resource.ResourceSlice), scope)
	}); err != nil {
		return err
	}
	if err := s.AddGeneratedConversionFunc((*resource.ResourceSlice)(nil), (*resourcev1beta1.ResourceSlice)(nil), func(a, b interface{}, scope conversion.Scope) error {
		return Convert_resource_ResourceSlice_To_v1beta1_ResourceSlice(a.(*resource.ResourceSlice), b.(*resourcev1beta1.ResourceSlice), scope)
	}); err != nil {
		return err
	}
	if err := s.AddGeneratedConversionFunc((*resourcev1beta1.ResourceSliceList)(nil), (*resource.ResourceSliceList)(nil), func(a, b interface{}, scope conversion.Scope) error {
		return Convert_v1beta1_ResourceSliceList_To_resource_ResourceSliceList(a.(*resourcev1beta1.ResourceSliceList), b.(*resource.ResourceSliceList), scope)
	}); err != nil {
		return err
	}
	if err := s.AddGeneratedConversionFunc((*resource.ResourceSliceList)(nil), (*resourcev1beta1.ResourceSliceList)(nil), func(a, b interface{}, scope conversion.Scope) error {
		return Convert_resource_ResourceSliceList_To_v1beta1_ResourceSliceList(a.(*resource.ResourceSliceList), b.(*resourcev1beta1.ResourceSliceList), scope)
	}); err != nil {
		return err
	}
	if err := s.AddConversionFunc((*resource.DeviceRequest)(nil), (*resourcev1beta1.DeviceRequest)(nil), func(a, b interface{}, scope conversion.Scope) error {
		return Convert_resource_DeviceRequest_To_v1beta1_DeviceRequest(a.(*resource.DeviceRequest), b.(*resourcev1beta1.DeviceRequest), scope)
	}); err != nil {
		return err
	}
	if err := s.AddConversionFunc((*resource.Device)(nil), (*resourcev1beta1.Device)(nil), func(a, b interface{}, scope conversion.Scope) error {
		return Convert_resource_Device_To_v1beta1_Device(a.(*resource.Device), b.(*resourcev1beta1.Device), scope)
	}); err != nil {
		return err
	}
	if err := s.AddConversionFunc((*resource.ResourceSliceSpec)(nil), (*resourcev1beta1.ResourceSliceSpec)(nil), func(a, b interface{}, scope conversion.Scope) error {
		return Convert_resource_ResourceSliceSpec_To_v1beta1_ResourceSliceSpec(a.(*resource.ResourceSliceSpec), b.(*resourcev1beta1.ResourceSliceSpec), scope)
	}); err != nil {
		return err
	}
	if err := s.AddConversionFunc((*resourcev1beta1.DeviceRequest)(nil), (*resource.DeviceRequest)(nil), func(a, b interface{}, scope conversion.Scope) error {
		return Convert_v1beta1_DeviceRequest_To_resource_DeviceRequest(a.(*resourcev1beta1.DeviceRequest), b.(*resource.DeviceRequest), scope)
	}); err != nil {
		return err
	}
	if err := s.AddConversionFunc((*resourcev1beta1.Device)(nil), (*resource.Device)(nil), func(a, b interface{}, scope conversion.Scope) error {
		return Convert_v1beta1_Device_To_resource_Device(a.(*resourcev1beta1.Device), b.(*resource.Device), scope)
	}); err != nil {
		return err
	}
	if err := s.AddConversionFunc((*resourcev1beta1.ResourceSliceSpec)(nil), (*resource.ResourceSliceSpec)(nil), func(a, b interface{}, scope conversion.Scope) error {
		return Convert_v1beta1_ResourceSliceSpec_To_resource_ResourceSliceSpec(a.(*resourcev1beta1.ResourceSliceSpec), b.(*resource.ResourceSliceSpec), scope)
	}); err != nil {
		return err
	}
	return nil
}

func autoConvert_v1beta1_AllocatedDeviceStatus_To_resource_AllocatedDeviceStatus(in *resourcev1beta1.AllocatedDeviceStatus, out *resource.AllocatedDeviceStatus, s conversion.Scope) error {
	out.Driver = in.Driver
	out.Pool = in.Pool
	out.Device = in.Device
	out.ShareID = (*string)(unsafe.Pointer(in.ShareID))
	out.Conditions = *(*[]v1.Condition)(unsafe.Pointer(&in.Conditions))
	out.Data = (*runtime.RawExtension)(unsafe.Pointer(in.Data))
	out.NetworkData = (*resource.NetworkDeviceData)(unsafe.Pointer(in.NetworkData))
	return nil
}

// Convert_v1beta1_AllocatedDeviceStatus_To_resource_AllocatedDeviceStatus is an autogenerated conversion function.
func Convert_v1beta1_AllocatedDeviceStatus_To_resource_AllocatedDeviceStatus(in *resourcev1beta1.AllocatedDeviceStatus, out *resource.AllocatedDeviceStatus, s conversion.Scope) error {
	return autoConvert_v1beta1_AllocatedDeviceStatus_To_resource_AllocatedDeviceStatus(in, out, s)
}

func autoConvert_resource_AllocatedDeviceStatus_To_v1beta1_AllocatedDeviceStatus(in *resource.AllocatedDeviceStatus, out *resourcev1beta1.AllocatedDeviceStatus, s conversion.Scope) error {
	out.Driver = in.Driver
	out.Pool = in.Pool
	out.Device = in.Device
	out.ShareID = (*string)(unsafe.Pointer(in.ShareID))
	out.Conditions = *(*[]v1.Condition)(unsafe.Pointer(&in.Conditions))
	out.Data = (*runtime.RawExtension)(unsafe.Pointer(in.Data))
	out.NetworkData = (*resourcev1beta1.NetworkDeviceData)(unsafe.Pointer(in.NetworkData))
	return nil
}

// Convert_resource_AllocatedDeviceStatus_To_v1beta1_AllocatedDeviceStatus is an autogenerated conversion function.
func Convert_resource_AllocatedDeviceStatus_To_v1beta1_AllocatedDeviceStatus(in *resource.AllocatedDeviceStatus, out *resourcev1beta1.AllocatedDeviceStatus, s conversion.Scope) error {
	return autoConvert_resource_AllocatedDeviceStatus_To_v1beta1_AllocatedDeviceStatus(in, out, s)
}

func autoConvert_v1beta1_AllocationResult_To_resource_AllocationResult(in *resourcev1beta1.AllocationResult, out *resource.AllocationResult, s conversion.Scope) error {
	if err := Convert_v1beta1_DeviceAllocationResult_To_resource_DeviceAllocationResult(&in.Devices, &out.Devices, s); err != nil {
		return err
	}
	out.NodeSelector = (*core.NodeSelector)(unsafe.Pointer(in.NodeSelector))
	out.AllocationTimestamp = (*v1.Time)(unsafe.Pointer(in.AllocationTimestamp))
	return nil
}

// Convert_v1beta1_AllocationResult_To_resource_AllocationResult is an autogenerated conversion function.
func Convert_v1beta1_AllocationResult_To_resource_AllocationResult(in *resourcev1beta1.AllocationResult, out *resource.AllocationResult, s conversion.Scope) error {
	return autoConvert_v1beta1_AllocationResult_To_resource_AllocationResult(in, out, s)
}

func autoConvert_resource_AllocationResult_To_v1beta1_AllocationResult(in *resource.AllocationResult, out *resourcev1beta1.AllocationResult, s conversion.Scope) error {
	if err := Convert_resource_DeviceAllocationResult_To_v1beta1_DeviceAllocationResult(&in.Devices, &out.Devices, s); err != nil {
		return err
	}
	out.NodeSelector = (*corev1.NodeSelector)(unsafe.Pointer(in.NodeSelector))
	out.AllocationTimestamp = (*v1.Time)(unsafe.Pointer(in.AllocationTimestamp))
	return nil
}

// Convert_resource_AllocationResult_To_v1beta1_AllocationResult is an autogenerated conversion function.
func Convert_resource_AllocationResult_To_v1beta1_AllocationResult(in *resource.AllocationResult, out *resourcev1beta1.AllocationResult, s conversion.Scope) error {
	return autoConvert_resource_AllocationResult_To_v1beta1_AllocationResult(in, out, s)
}

func autoConvert_v1beta1_CELDeviceSelector_To_resource_CELDeviceSelector(in *resourcev1beta1.CELDeviceSelector, out *resource.CELDeviceSelector, s conversion.Scope) error {
	out.Expression = in.Expression
	return nil
}

// Convert_v1beta1_CELDeviceSelector_To_resource_CELDeviceSelector is an autogenerated conversion function.
func Convert_v1beta1_CELDeviceSelector_To_resource_CELDeviceSelector(in *resourcev1beta1.CELDeviceSelector, out *resource.CELDeviceSelector, s conversion.Scope) error {
	return autoConvert_v1beta1_CELDeviceSelector_To_resource_CELDeviceSelector(in, out, s)
}

func autoConvert_resource_CELDeviceSelector_To_v1beta1_CELDeviceSelector(in *resource.CELDeviceSelector, out *resourcev1beta1.CELDeviceSelector, s conversion.Scope) error {
	out.Expression = in.Expression
	return nil
}

// Convert_resource_CELDeviceSelector_To_v1beta1_CELDeviceSelector is an autogenerated conversion function.
func Convert_resource_CELDeviceSelector_To_v1beta1_CELDeviceSelector(in *resource.CELDeviceSelector, out *resourcev1beta1.CELDeviceSelector, s conversion.Scope) error {
	return autoConvert_resource_CELDeviceSelector_To_v1beta1_CELDeviceSelector(in, out, s)
}

func autoConvert_v1beta1_CapacityRequestPolicy_To_resource_CapacityRequestPolicy(in *resourcev1beta1.CapacityRequestPolicy, out *resource.CapacityRequestPolicy, s conversion.Scope) error {
	out.Default = (*apiresource.Quantity)(unsafe.Pointer(in.Default))
	out.ValidValues = *(*[]apiresource.Quantity)(unsafe.Pointer(&in.ValidValues))
	out.ValidRange = (*resource.CapacityRequestPolicyRange)(unsafe.Pointer(in.ValidRange))
	return nil
}

// Convert_v1beta1_CapacityRequestPolicy_To_resource_CapacityRequestPolicy is an autogenerated conversion function.
func Convert_v1beta1_CapacityRequestPolicy_To_resource_CapacityRequestPolicy(in *resourcev1beta1.CapacityRequestPolicy, out *resource.CapacityRequestPolicy, s conversion.Scope) error {
	return autoConvert_v1beta1_CapacityRequestPolicy_To_resource_CapacityRequestPolicy(in, out, s)
}

func autoConvert_resource_CapacityRequestPolicy_To_v1beta1_CapacityRequestPolicy(in *resource.CapacityRequestPolicy, out *resourcev1beta1.CapacityRequestPolicy, s conversion.Scope) error {
	out.Default = (*apiresource.Quantity)(unsafe.Pointer(in.Default))
	out.ValidValues = *(*[]apiresource.Quantity)(unsafe.Pointer(&in.ValidValues))
	out.ValidRange = (*resourcev1beta1.CapacityRequestPolicyRange)(unsafe.Pointer(in.ValidRange))
	return nil
}

// Convert_resource_CapacityRequestPolicy_To_v1beta1_CapacityRequestPolicy is an autogenerated conversion function.
func Convert_resource_CapacityRequestPolicy_To_v1beta1_CapacityRequestPolicy(in *resource.CapacityRequestPolicy, out *resourcev1beta1.CapacityRequestPolicy, s conversion.Scope) error {
	return autoConvert_resource_CapacityRequestPolicy_To_v1beta1_CapacityRequestPolicy(in, out, s)
}

func autoConvert_v1beta1_CapacityRequestPolicyRange_To_resource_CapacityRequestPolicyRange(in *resourcev1beta1.CapacityRequestPolicyRange, out *resource.CapacityRequestPolicyRange, s conversion.Scope) error {
	out.Min = (*apiresource.Quantity)(unsafe.Pointer(in.Min))
	out.Max = (*apiresource.Quantity)(unsafe.Pointer(in.Max))
	out.Step = (*apiresource.Quantity)(unsafe.Pointer(in.Step))
	return nil
}

// Convert_v1beta1_CapacityRequestPolicyRange_To_resource_CapacityRequestPolicyRange is an autogenerated conversion function.
func Convert_v1beta1_CapacityRequestPolicyRange_To_resource_CapacityRequestPolicyRange(in *resourcev1beta1.CapacityRequestPolicyRange, out *resource.CapacityRequestPolicyRange, s conversion.Scope) error {
	return autoConvert_v1beta1_CapacityRequestPolicyRange_To_resource_CapacityRequestPolicyRange(in, out, s)
}

func autoConvert_resource_CapacityRequestPolicyRange_To_v1beta1_CapacityRequestPolicyRange(in *resource.CapacityRequestPolicyRange, out *resourcev1beta1.CapacityRequestPolicyRange, s conversion.Scope) error {
	out.Min = (*apiresource.Quantity)(unsafe.Pointer(in.Min))
	out.Max = (*apiresource.Quantity)(unsafe.Pointer(in.Max))
	out.Step = (*apiresource.Quantity)(unsafe.Pointer(in.Step))
	return nil
}

// Convert_resource_CapacityRequestPolicyRange_To_v1beta1_CapacityRequestPolicyRange is an autogenerated conversion function.
func Convert_resource_CapacityRequestPolicyRange_To_v1beta1_CapacityRequestPolicyRange(in *resource.CapacityRequestPolicyRange, out *resourcev1beta1.CapacityRequestPolicyRange, s conversion.Scope) error {
	return autoConvert_resource_CapacityRequestPolicyRange_To_v1beta1_CapacityRequestPolicyRange(in, out, s)
}

func autoConvert_v1beta1_CapacityRequirements_To_resource_CapacityRequirements(in *resourcev1beta1.CapacityRequirements, out *resource.CapacityRequirements, s conversion.Scope) error {
	out.Requests = *(*map[resource.QualifiedName]apiresource.Quantity)(unsafe.Pointer(&in.Requests))
	return nil
}

// Convert_v1beta1_CapacityRequirements_To_resource_CapacityRequirements is an autogenerated conversion function.
func Convert_v1beta1_CapacityRequirements_To_resource_CapacityRequirements(in *resourcev1beta1.CapacityRequirements, out *resource.CapacityRequirements, s conversion.Scope) error {
	return autoConvert_v1beta1_CapacityRequirements_To_resource_CapacityRequirements(in, out, s)
}

func autoConvert_resource_CapacityRequirements_To_v1beta1_CapacityRequirements(in *resource.CapacityRequirements, out *resourcev1beta1.CapacityRequirements, s conversion.Scope) error {
	out.Requests = *(*map[resourcev1beta1.QualifiedName]apiresource.Quantity)(unsafe.Pointer(&in.Requests))
	return nil
}

// Convert_resource_CapacityRequirements_To_v1beta1_CapacityRequirements is an autogenerated conversion function.
func Convert_resource_CapacityRequirements_To_v1beta1_CapacityRequirements(in *resource.CapacityRequirements, out *resourcev1beta1.CapacityRequirements, s conversion.Scope) error {
	return autoConvert_resource_CapacityRequirements_To_v1beta1_CapacityRequirements(in, out, s)
}

func autoConvert_v1beta1_Counter_To_resource_Counter(in *resourcev1beta1.Counter, out *resource.Counter, s conversion.Scope) error {
	out.Value = in.Value
	return nil
}

// Convert_v1beta1_Counter_To_resource_Counter is an autogenerated conversion function.
func Convert_v1beta1_Counter_To_resource_Counter(in *resourcev1beta1.Counter, out *resource.Counter, s conversion.Scope) error {
	return autoConvert_v1beta1_Counter_To_resource_Counter(in, out, s)
}

func autoConvert_resource_Counter_To_v1beta1_Counter(in *resource.Counter, out *resourcev1beta1.Counter, s conversion.Scope) error {
	out.Value = in.Value
	return nil
}

// Convert_resource_Counter_To_v1beta1_Counter is an autogenerated conversion function.
func Convert_resource_Counter_To_v1beta1_Counter(in *resource.Counter, out *resourcev1beta1.Counter, s conversion.Scope) error {
	return autoConvert_resource_Counter_To_v1beta1_Counter(in, out, s)
}

func autoConvert_v1beta1_CounterSet_To_resource_CounterSet(in *resourcev1beta1.CounterSet, out *resource.CounterSet, s conversion.Scope) error {
	out.Name = in.Name
	out.Counters = *(*map[string]resource.Counter)(unsafe.Pointer(&in.Counters))
	return nil
}

// Convert_v1beta1_CounterSet_To_resource_CounterSet is an autogenerated conversion function.
func Convert_v1beta1_CounterSet_To_resource_CounterSet(in *resourcev1beta1.CounterSet, out *resource.CounterSet, s conversion.Scope) error {
	return autoConvert_v1beta1_CounterSet_To_resource_CounterSet(in, out, s)
}

func autoConvert_resource_CounterSet_To_v1beta1_CounterSet(in *resource.CounterSet, out *resourcev1beta1.CounterSet, s conversion.Scope) error {
	out.Name = in.Name
	out.Counters = *(*map[string]resourcev1beta1.Counter)(unsafe.Pointer(&in.Counters))
	return nil
}

// Convert_resource_CounterSet_To_v1beta1_CounterSet is an autogenerated conversion function.
func Convert_resource_CounterSet_To_v1beta1_CounterSet(in *resource.CounterSet, out *resourcev1beta1.CounterSet, s conversion.Scope) error {
	return autoConvert_resource_CounterSet_To_v1beta1_CounterSet(in, out, s)
}

func autoConvert_v1beta1_Device_To_resource_Device(in *resourcev1beta1.Device, out *resource.Device, s conversion.Scope) error {
	out.Name = in.Name
	// WARNING: in.Basic requires manual conversion: does not exist in peer-type
	return nil
}

func autoConvert_resource_Device_To_v1beta1_Device(in *resource.Device, out *resourcev1beta1.Device, s conversion.Scope) error {
	out.Name = in.Name
	// WARNING: in.Attributes requires manual conversion: does not exist in peer-type
	// WARNING: in.Capacity requires manual conversion: does not exist in peer-type
	// WARNING: in.ConsumesCounters requires manual conversion: does not exist in peer-type
	// WARNING: in.NodeName requires manual conversion: does not exist in peer-type
	// WARNING: in.NodeSelector requires manual conversion: does not exist in peer-type
	// WARNING: in.AllNodes requires manual conversion: does not exist in peer-type
	// WARNING: in.Taints requires manual conversion: does not exist in peer-type
	// WARNING: in.BindsToNode requires manual conversion: does not exist in peer-type
	// WARNING: in.BindingConditions requires manual conversion: does not exist in peer-type
	// WARNING: in.BindingFailureConditions requires manual conversion: does not exist in peer-type
	// WARNING: in.AllowMultipleAllocations requires manual conversion: does not exist in peer-type
	// WARNING: in.NodeAllocatableResourceMappings requires manual conversion: does not exist in peer-type
	return nil
}

func autoConvert_v1beta1_DeviceAllocationConfiguration_To_resource_DeviceAllocationConfiguration(in *resourcev1beta1.DeviceAllocationConfiguration, out *resource.DeviceAllocationConfiguration, s conversion.Scope) error {
	out.Source = resource.AllocationConfigSource(in.Source)
	out.Requests = *(*[]string)(unsafe.Pointer(&in.Requests))
	if err := Convert_v1beta1_DeviceConfiguration_To_resource_DeviceConfiguration(&in.DeviceConfiguration, &out.DeviceConfiguration, s); err != nil {
		return err
	}
	return nil
}

// Convert_v1beta1_DeviceAllocationConfiguration_To_resource_DeviceAllocationConfiguration is an autogenerated conversion function.
func Convert_v1beta1_DeviceAllocationConfiguration_To_resource_DeviceAllocationConfiguration(in *resourcev1beta1.DeviceAllocationConfiguration, out *resource.DeviceAllocationConfiguration, s conversion.Scope) error {
	return autoConvert_v1beta1_DeviceAllocationConfiguration_To_resource_DeviceAllocationConfiguration(in, out, s)
}

func autoConvert_resource_DeviceAllocationConfiguration_To_v1beta1_DeviceAllocationConfiguration(in *resource.DeviceAllocationConfiguration, out *resourcev1beta1.DeviceAllocationConfiguration, s conversion.Scope) error {
	out.Source = resourcev1beta1.AllocationConfigSource(in.Source)
	out.Requests = *(*[]string)(unsafe.Pointer(&in.Requests))
	if err := Convert_resource_DeviceConfiguration_To_v1beta1_DeviceConfiguration(&in.DeviceConfiguration, &out.DeviceConfiguration, s); err != nil {
		return err
	}
	return nil
}

// Convert_resource_DeviceAllocationConfiguration_To_v1beta1_DeviceAllocationConfiguration is an autogenerated conversion function.
func Convert_resource_DeviceAllocationConfiguration_To_v1beta1_DeviceAllocationConfiguration(in *resource.DeviceAllocationConfiguration, out *resourcev1beta1.DeviceAllocationConfiguration, s conversion.Scope) error {
	return autoConvert_resource_DeviceAllocationConfiguration_To_v1beta1_DeviceAllocationConfiguration(in, out, s)
}

func autoConvert_v1beta1_DeviceAllocationResult_To_resource_DeviceAllocationResult(in *resourcev1beta1.DeviceAllocationResult, out *resource.DeviceAllocationResult, s conversion.Scope) error {
	out.Results = *(*[]resource.DeviceRequestAllocationResult)(unsafe.Pointer(&in.Results))
	out.Config = *(*[]resource.DeviceAllocationConfiguration)(unsafe.Pointer(&in.Config))
	return nil
}

// Convert_v1beta1_DeviceAllocationResult_To_resource_DeviceAllocationResult is an autogenerated conversion function.
func Convert_v1beta1_DeviceAllocationResult_To_resource_DeviceAllocationResult(in *resourcev1beta1.DeviceAllocationResult, out *resource.DeviceAllocationResult, s conversion.Scope) error {
	return autoConvert_v1beta1_DeviceAllocationResult_To_resource_DeviceAllocationResult(in, out, s)
}

func autoConvert_resource_DeviceAllocationResult_To_v1beta1_DeviceAllocationResult(in *resource.DeviceAllocationResult, out *resourcev1beta1.DeviceAllocationResult, s conversion.Scope) error {
	out.Results = *(*[]resourcev1beta1.DeviceRequestAllocationResult)(unsafe.Pointer(&in.Results))
	out.Config = *(*[]resourcev1beta1.DeviceAllocationConfiguration)(unsafe.Pointer(&in.Config))
	return nil
}

// Convert_resource_DeviceAllocationResult_To_v1beta1_DeviceAllocationResult is an autogenerated conversion function.
func Convert_resource_DeviceAllocationResult_To_v1beta1_DeviceAllocationResult(in *resource.DeviceAllocationResult, out *resourcev1beta1.DeviceAllocationResult, s conversion.Scope) error {
	return autoConvert_resource_DeviceAllocationResult_To_v1beta1_DeviceAllocationResult(in, out, s)
}

func autoConvert_v1beta1_DeviceAttribute_To_resource_DeviceAttribute(in *resourcev1beta1.DeviceAttribute, out *resource.DeviceAttribute, s conversion.Scope) error {
	out.IntValue = (*int64)(unsafe.Pointer(in.IntValue))
	out.BoolValue = (*bool)(unsafe.Pointer(in.BoolValue))
	out.StringValue = (*string)(unsafe.Pointer(in.StringValue))
	out.VersionValue = (*string)(unsafe.Pointer(in.VersionValue))
	out.IntValues = *(*[]int64)(unsafe.Pointer(&in.IntValues))
	out.BoolValues = *(*[]bool)(unsafe.Pointer(&in.BoolValues))
	out.StringValues = *(*[]string)(unsafe.Pointer(&in.StringValues))
	out.VersionValues = *(*[]string)(unsafe.Pointer(&in.VersionValues))
	return nil
}

// Convert_v1beta1_DeviceAttribute_To_resource_DeviceAttribute is an autogenerated conversion function.
func Convert_v1beta1_DeviceAttribute_To_resource_DeviceAttribute(in *resourcev1beta1.DeviceAttribute, out *resource.DeviceAttribute, s conversion.Scope) error {
	return autoConvert_v1beta1_DeviceAttribute_To_resource_DeviceAttribute(in, out, s)
}

func autoConvert_resource_DeviceAttribute_To_v1beta1_DeviceAttribute(in *resource.DeviceAttribute, out *resourcev1beta1.DeviceAttribute, s conversion.Scope) error {
	out.IntValue = (*int64)(unsafe.Pointer(in.IntValue))
	out.BoolValue = (*bool)(unsafe.Pointer(in.BoolValue))
	out.StringValue = (*string)(unsafe.Pointer(in.StringValue))
	out.VersionValue = (*string)(unsafe.Pointer(in.VersionValue))
	out.IntValues = *(*[]int64)(unsafe.Pointer(&in.IntValues))
	out.BoolValues = *(*[]bool)(unsafe.Pointer(&in.BoolValues))
	out.StringValues = *(*[]string)(unsafe.Pointer(&in.StringValues))
	out.VersionValues = *(*[]string)(unsafe.Pointer(&in.VersionValues))
	return nil
}

// Convert_resource_DeviceAttribute_To_v1beta1_DeviceAttribute is an autogenerated conversion function.
func Convert_resource_DeviceAttribute_To_v1beta1_DeviceAttribute(in *resource.DeviceAttribute, out *resourcev1beta1.DeviceAttribute, s conversion.Scope) error {
	return autoConvert_resource_DeviceAttribute_To_v1beta1_DeviceAttribute(in, out, s)
}

func autoConvert_v1beta1_DeviceCapacity_To_resource_DeviceCapacity(in *resourcev1beta1.DeviceCapacity, out *resource.DeviceCapacity, s conversion.Scope) error {
	out.Value = in.Value
	out.RequestPolicy = (*resource.CapacityRequestPolicy)(unsafe.Pointer(in.RequestPolicy))
	return nil
}

// Convert_v1beta1_DeviceCapacity_To_resource_DeviceCapacity is an autogenerated conversion function.
func Convert_v1beta1_DeviceCapacity_To_resource_DeviceCapacity(in *resourcev1beta1.DeviceCapacity, out *resource.DeviceCapacity, s conversion.Scope) error {
	return autoConvert_v1beta1_DeviceCapacity_To_resource_DeviceCapacity(in, out, s)
}

func autoConvert_resource_DeviceCapacity_To_v1beta1_DeviceCapacity(in *resource.DeviceCapacity, out *resourcev1beta1.DeviceCapacity, s conversion.Scope) error {
	out.Value = in.Value
	out.RequestPolicy = (*resourcev1beta1.CapacityRequestPolicy)(unsafe.Pointer(in.RequestPolicy))
	return nil
}

// Convert_resource_DeviceCapacity_To_v1beta1_DeviceCapacity is an autogenerated conversion function.
func Convert_resource_DeviceCapacity_To_v1beta1_DeviceCapacity(in *resource.DeviceCapacity, out *resourcev1beta1.DeviceCapacity, s conversion.Scope) error {
	return autoConvert_resource_DeviceCapacity_To_v1beta1_DeviceCapacity(in, out, s)
}

func autoConvert_v1beta1_DeviceClaim_To_resource_DeviceClaim(in *resourcev1beta1.DeviceClaim, out *resource.DeviceClaim, s conversion.Scope) error {
	if in.Requests != nil {
		in, out := &in.Requests, &out.Requests
		*out = make([]resource.DeviceRequest, len(*in))
		for i := range *in {
			if err := Convert_v1beta1_DeviceRequest_To_resource_DeviceRequest(&(*in)[i], &(*out)[i], s); err != nil {
				return err
			}
		}
	} else {
		out.Requests = nil
	}
	out.Constraints = *(*[]resource.DeviceConstraint)(unsafe.Pointer(&in.Constraints))
	out.Config = *(*[]resource.DeviceClaimConfiguration)(unsafe.Pointer(&in.Config))
	return nil
}

// Convert_v1beta1_DeviceClaim_To_resource_DeviceClaim is an autogenerated conversion function.
func Convert_v1beta1_DeviceClaim_To_resource_DeviceClaim(in *resourcev1beta1.DeviceClaim, out *resource.DeviceClaim, s conversion.Scope) error {
	return autoConvert_v1beta1_DeviceClaim_To_resource_DeviceClaim(in, out, s)
}

func autoConvert_resource_DeviceClaim_To_v1beta1_DeviceClaim(in *resource.DeviceClaim, out *resourcev1beta1.DeviceClaim, s conversion.Scope) error {
	if in.Requests != nil {
		in, out := &in.Requests, &out.Requests
		*out = make([]resourcev1beta1.DeviceRequest, len(*in))
		for i := range *in {
			if err := Convert_resource_DeviceRequest_To_v1beta1_DeviceRequest(&(*in)[i], &(*out)[i], s); err != nil {
				return err
			}
		}
	} else {
		out.Requests = nil
	}
	out.Constraints = *(*[]resourcev1beta1.DeviceConstraint)(unsafe.Pointer(&in.Constraints))
	out.Config = *(*[]resourcev1beta1.DeviceClaimConfiguration)(unsafe.Pointer(&in.Config))
	return nil
}

// Convert_resource_DeviceClaim_To_v1beta1_DeviceClaim is an autogenerated conversion function.
func Convert_resource_DeviceClaim_To_v1beta1_DeviceClaim(in *resource.DeviceClaim, out *resourcev1beta1.DeviceClaim, s conversion.Scope) error {
	return autoConvert_resource_DeviceClaim_To_v1beta1_DeviceClaim(in, out, s)
}

func autoConvert_v1beta1_DeviceClaimConfiguration_To_resource_DeviceClaimConfiguration(in *resourcev1beta1.DeviceClaimConfiguration, out *resource.DeviceClaimConfiguration, s conversion.Scope) error {
	out.Requests = *(*[]string)(unsafe.Pointer(&in.Requests))
	if err := Convert_v1beta1_DeviceConfiguration_To_resource_DeviceConfiguration(&in.DeviceConfiguration, &out.DeviceConfiguration, s); err != nil {
		return err
	}
	return nil
}

// Convert_v1beta1_DeviceClaimConfiguration_To_resource_DeviceClaimConfiguration is an autogenerated conversion function.
func Convert_v1beta1_DeviceClaimConfiguration_To_resource_DeviceClaimConfiguration(in *resourcev1beta1.DeviceClaimConfiguration, out *resource.DeviceClaimConfiguration, s conversion.Scope) error {
	return autoConvert_v1beta1_DeviceClaimConfiguration_To_resource_DeviceClaimConfiguration(in, out, s)
}

func autoConvert_resource_DeviceClaimConfiguration_To_v1beta1_DeviceClaimConfiguration(in *resource.DeviceClaimConfiguration, out *resourcev1beta1.DeviceClaimConfiguration, s conversion.Scope) error {
	out.Requests = *(*[]string)(unsafe.Pointer(&in.Requests))
	if err := Convert_resource_DeviceConfiguration_To_v1beta1_DeviceConfiguration(&in.DeviceConfiguration, &out.DeviceConfiguration, s); err != nil {
		return err
	}
	return nil
}

// Convert_resource_DeviceClaimConfiguration_To_v1beta1_DeviceClaimConfiguration is an autogenerated conversion function.
func Convert_resource_DeviceClaimConfiguration_To_v1beta1_DeviceClaimConfiguration(in *resource.DeviceClaimConfiguration, out *resourcev1beta1.DeviceClaimConfiguration, s conversion.Scope) error {
	return autoConvert_resource_DeviceClaimConfiguration_To_v1beta1_DeviceClaimConfiguration(in, out, s)
}

func autoConvert_v1beta1_DeviceClass_To_resource_DeviceClass(in *resourcev1beta1.DeviceClass, out *resource.DeviceClass, s conversion.Scope) error {
	out.ObjectMeta = in.ObjectMeta
	if err := Convert_v1beta1_DeviceClassSpec_To_resource_DeviceClassSpec(&in.Spec, &out.Spec, s); err != nil {
		return err
	}
	return nil
}

// Convert_v1beta1_DeviceClass_To_resource_DeviceClass is an autogenerated conversion function.
func Convert_v1beta1_DeviceClass_To_resource_DeviceClass(in *resourcev1beta1.DeviceClass, out *resource.DeviceClass, s conversion.Scope) error {
	return autoConvert_v1beta1_DeviceClass_To_resource_DeviceClass(in, out, s)
}

func autoConvert_resource_DeviceClass_To_v1beta1_DeviceClass(in *resource.DeviceClass, out *resourcev1beta1.DeviceClass, s conversion.Scope) error {
	out.ObjectMeta = in.ObjectMeta
	if err := Convert_resource_DeviceClassSpec_To_v1beta1_DeviceClassSpec(&in.Spec, &out.Spec, s); err != nil {
		return err
	}
	return nil
}

// Convert_resource_DeviceClass_To_v1beta1_DeviceClass is an autogenerated conversion function.
func Convert_resource_DeviceClass_To_v1beta1_DeviceClass(in *resource.DeviceClass, out *resourcev1beta1.DeviceClass, s conversion.Scope) error {
	return autoConvert_resource_DeviceClass_To_v1beta1_DeviceClass(in, out, s)
}

func autoConvert_v1beta1_DeviceClassConfiguration_To_resource_DeviceClassConfiguration(in *resourcev1beta1.DeviceClassConfiguration, out *resource.DeviceClassConfiguration, s conversion.Scope) error {
	if err := Convert_v1beta1_DeviceConfiguration_To_resource_DeviceConfiguration(&in.DeviceConfiguration, &out.DeviceConfiguration, s); err != nil {
		return err
	}
	return nil
}

// Convert_v1beta1_DeviceClassConfiguration_To_resource_DeviceClassConfiguration is an autogenerated conversion function.
func Convert_v1beta1_DeviceClassConfiguration_To_resource_DeviceClassConfiguration(in *resourcev1beta1.DeviceClassConfiguration, out *resource.DeviceClassConfiguration, s conversion.Scope) error {
	return autoConvert_v1beta1_DeviceClassConfiguration_To_resource_DeviceClassConfiguration(in, out, s)
}

func autoConvert_resource_DeviceClassConfiguration_To_v1beta1_DeviceClassConfiguration(in *resource.DeviceClassConfiguration, out *resourcev1beta1.DeviceClassConfiguration, s conversion.Scope) error {
	if err := Convert_resource_DeviceConfiguration_To_v1beta1_DeviceConfiguration(&in.DeviceConfiguration, &out.DeviceConfiguration, s); err != nil {
		return err
	}
	return nil
}

// Convert_resource_DeviceClassConfiguration_To_v1beta1_DeviceClassConfiguration is an autogenerated conversion function.
func Convert_resource_DeviceClassConfiguration_To_v1beta1_DeviceClassConfiguration(in *resource.DeviceClassConfiguration, out *resourcev1beta1.DeviceClassConfiguration, s conversion.Scope) error {
	return autoConvert_resource_DeviceClassConfiguration_To_v1beta1_DeviceClassConfiguration(in, out, s)
}

func autoConvert_v1beta1_DeviceClassList_To_resource_DeviceClassList(in *resourcev1beta1.DeviceClassList, out *resource.DeviceClassList, s conversion.Scope) error {
	out.ListMeta = in.ListMeta
	out.Items = *(*[]resource.DeviceClass)(unsafe.Pointer(&in.Items))
	return nil
}

// Convert_v1beta1_DeviceClassList_To_resource_DeviceClassList is an autogenerated conversion function.
func Convert_v1beta1_DeviceClassList_To_resource_DeviceClassList(in *resourcev1beta1.DeviceClassList, out *resource.DeviceClassList, s conversion.Scope) error {
	return autoConvert_v1beta1_DeviceClassList_To_resource_DeviceClassList(in, out, s)
}

func autoConvert_resource_DeviceClassList_To_v1beta1_DeviceClassList(in *resource.DeviceClassList, out *resourcev1beta1.DeviceClassList, s conversion.Scope) error {
	out.ListMeta = in.ListMeta
	out.Items = *(*[]resourcev1beta1.DeviceClass)(unsafe.Pointer(&in.Items))
	return nil
}

// Convert_resource_DeviceClassList_To_v1beta1_DeviceClassList is an autogenerated conversion function.
func Convert_resource_DeviceClassList_To_v1beta1_DeviceClassList(in *resource.DeviceClassList, out *resourcev1beta1.DeviceClassList, s conversion.Scope) error {
	return autoConvert_resource_DeviceClassList_To_v1beta1_DeviceClassList(in, out, s)
}

func autoConvert_v1beta1_DeviceClassSpec_To_resource_DeviceClassSpec(in *resourcev1beta1.DeviceClassSpec, out *resource.DeviceClassSpec, s conversion.Scope) error {
	out.Selectors = *(*[]resource.DeviceSelector)(unsafe.Pointer(&in.Selectors))
	out.Config = *(*[]resource.DeviceClassConfiguration)(unsafe.Pointer(&in.Config))
	out.ExtendedResourceName = (*string)(unsafe.Pointer(in.ExtendedResourceName))
	return nil
}

// Convert_v1beta1_DeviceClassSpec_To_resource_DeviceClassSpec is an autogenerated conversion function.
func Convert_v1beta1_DeviceClassSpec_To_resource_DeviceClassSpec(in *resourcev1beta1.DeviceClassSpec, out *resource.DeviceClassSpec, s conversion.Scope) error {
	return autoConvert_v1beta1_DeviceClassSpec_To_resource_DeviceClassSpec(in, out, s)
}

func autoConvert_resource_DeviceClassSpec_To_v1beta1_DeviceClassSpec(in *resource.DeviceClassSpec, out *resourcev1beta1.DeviceClassSpec, s conversion.Scope) error {
	out.Selectors = *(*[]resourcev1beta1.DeviceSelector)(unsafe.Pointer(&in.Selectors))
	out.Config = *(*[]resourcev1beta1.DeviceClassConfiguration)(unsafe.Pointer(&in.Config))
	out.ExtendedResourceName = (*string)(unsafe.Pointer(in.ExtendedResourceName))
	return nil
}

// Convert_resource_DeviceClassSpec_To_v1beta1_DeviceClassSpec is an autogenerated conversion function.
func Convert_resource_DeviceClassSpec_To_v1beta1_DeviceClassSpec(in *resource.DeviceClassSpec, out *resourcev1beta1.DeviceClassSpec, s conversion.Scope) error {
	return autoConvert_resource_DeviceClassSpec_To_v1beta1_DeviceClassSpec(in, out, s)
}

func autoConvert_v1beta1_DeviceConfiguration_To_resource_DeviceConfiguration(in *resourcev1beta1.DeviceConfiguration, out *resource.DeviceConfiguration, s conversion.Scope) error {
	out.Opaque = (*resource.OpaqueDeviceConfiguration)(unsafe.Pointer(in.Opaque))
	return nil
}

// Convert_v1beta1_DeviceConfiguration_To_resource_DeviceConfiguration is an autogenerated conversion function.
func Convert_v1beta1_DeviceConfiguration_To_resource_DeviceConfiguration(in *resourcev1beta1.DeviceConfiguration, out *resource.DeviceConfiguration, s conversion.Scope) error {
	return autoConvert_v1beta1_DeviceConfiguration_To_resource_DeviceConfiguration(in, out, s)
}

func autoConvert_resource_DeviceConfiguration_To_v1beta1_DeviceConfiguration(in *resource.DeviceConfiguration, out *resourcev1beta1.DeviceConfiguration, s conversion.Scope) error {
	out.Opaque = (*resourcev1beta1.OpaqueDeviceConfiguration)(unsafe.Pointer(in.Opaque))
	return nil
}

// Convert_resource_DeviceConfiguration_To_v1beta1_DeviceConfiguration is an autogenerated conversion function.
func Convert_resource_DeviceConfiguration_To_v1beta1_DeviceConfiguration(in *resource.DeviceConfiguration, out *resourcev1beta1.DeviceConfiguration, s conversion.Scope) error {
	return autoConvert_resource_DeviceConfiguration_To_v1beta1_DeviceConfiguration(in, out, s)
}

func autoConvert_v1beta1_DeviceConstraint_To_resource_DeviceConstraint(in *resourcev1beta1.DeviceConstraint, out *resource.DeviceConstraint, s conversion.Scope) error {
	out.Requests = *(*[]string)(unsafe.Pointer(&in.Requests))
	out.MatchAttribute = (*resource.FullyQualifiedName)(unsafe.Pointer(in.MatchAttribute))
	out.DistinctAttribute = (*resource.FullyQualifiedName)(unsafe.Pointer(in.DistinctAttribute))
	return nil
}

// Convert_v1beta1_DeviceConstraint_To_resource_DeviceConstraint is an autogenerated conversion function.
func Convert_v1beta1_DeviceConstraint_To_resource_DeviceConstraint(in *resourcev1beta1.DeviceConstraint, out *resource.DeviceConstraint, s conversion.Scope) error {
	return autoConvert_v1beta1_DeviceConstraint_To_resource_DeviceConstraint(in, out, s)
}

func autoConvert_resource_DeviceConstraint_To_v1beta1_DeviceConstraint(in *resource.DeviceConstraint, out *resourcev1beta1.DeviceConstraint, s conversion.Scope) error {
	out.Requests = *(*[]string)(unsafe.Pointer(&in.Requests))
	out.MatchAttribute = (*resourcev1beta1.FullyQualifiedName)(unsafe.Pointer(in.MatchAttribute))
	out.DistinctAttribute = (*resourcev1beta1.FullyQualifiedName)(unsafe.Pointer(in.DistinctAttribute))
	return nil
}

// Convert_resource_DeviceConstraint_To_v1beta1_DeviceConstraint is an autogenerated conversion function.
func Convert_resource_DeviceConstraint_To_v1beta1_DeviceConstraint(in *resource.DeviceConstraint, out *resourcev1beta1.DeviceConstraint, s conversion.Scope) error {
	return autoConvert_resource_DeviceConstraint_To_v1beta1_DeviceConstraint(in, out, s)
}

func autoConvert_v1beta1_DeviceCounterConsumption_To_resource_DeviceCounterConsumption(in *resourcev1beta1.DeviceCounterConsumption, out *resource.DeviceCounterConsumption, s conversion.Scope) error {
	out.CounterSet = in.CounterSet
	out.Counters = *(*map[string]resource.Counter)(unsafe.Pointer(&in.Counters))
	return nil
}

// Convert_v1beta1_DeviceCounterConsumption_To_resource_DeviceCounterConsumption is an autogenerated conversion function.
func Convert_v1beta1_DeviceCounterConsumption_To_resource_DeviceCounterConsumption(in *resourcev1beta1.DeviceCounterConsumption, out *resource.DeviceCounterConsumption, s conversion.Scope) error {
	return autoConvert_v1beta1_DeviceCounterConsumption_To_resource_DeviceCounterConsumption(in, out, s)
}

func autoConvert_resource_DeviceCounterConsumption_To_v1beta1_DeviceCounterConsumption(in *resource.DeviceCounterConsumption, out *resourcev1beta1.DeviceCounterConsumption, s conversion.Scope) error {
	out.CounterSet = in.CounterSet
	out.Counters = *(*map[string]resourcev1beta1.Counter)(unsafe.Pointer(&in.Counters))
	return nil
}

// Convert_resource_DeviceCounterConsumption_To_v1beta1_DeviceCounterConsumption is an autogenerated conversion function.
func Convert_resource_DeviceCounterConsumption_To_v1beta1_DeviceCounterConsumption(in *resource.DeviceCounterConsumption, out *resourcev1beta1.DeviceCounterConsumption, s conversion.Scope) error {
	return autoConvert_resource_DeviceCounterConsumption_To_v1beta1_DeviceCounterConsumption(in, out, s)
}

func autoConvert_v1beta1_DeviceRequest_To_resource_DeviceRequest(in *resourcev1beta1.DeviceRequest, out *resource.DeviceRequest, s conversion.Scope) error {
	out.Name = in.Name
	// WARNING: in.DeviceClassName requires manual conversion: does not exist in peer-type
	// WARNING: in.Selectors requires manual conversion: does not exist in peer-type
	// WARNING: in.AllocationMode requires manual conversion: does not exist in peer-type
	// WARNING: in.Count requires manual conversion: does not exist in peer-type
	// WARNING: in.AdminAccess requires manual conversion: does not exist in peer-type
	out.FirstAvailable = *(*[]resource.DeviceSubRequest)(unsafe.Pointer(&in.FirstAvailable))
	// WARNING: in.Tolerations requires manual conversion: does not exist in peer-type
	// WARNING: in.Capacity requires manual conversion: does not exist in peer-type
	return nil
}

func autoConvert_resource_DeviceRequest_To_v1beta1_DeviceRequest(in *resource.DeviceRequest, out *resourcev1beta1.DeviceRequest, s conversion.Scope) error {
	out.Name = in.Name
	// WARNING: in.Exactly requires manual conversion: does not exist in peer-type
	out.FirstAvailable = *(*[]resourcev1beta1.DeviceSubRequest)(unsafe.Pointer(&in.FirstAvailable))
	return nil
}

func autoConvert_v1beta1_DeviceRequestAllocationResult_To_resource_DeviceRequestAllocationResult(in *resourcev1beta1.DeviceRequestAllocationResult, out *resource.DeviceRequestAllocationResult, s conversion.Scope) error {
	out.Request = in.Request
	out.Driver = in.Driver
	out.Pool = in.Pool
	out.Device = in.Device
	out.AdminAccess = (*bool)(unsafe.Pointer(in.AdminAccess))
	out.Tolerations = *(*[]resource.DeviceToleration)(unsafe.Pointer(&in.Tolerations))
	out.BindingConditions = *(*[]string)(unsafe.Pointer(&in.BindingConditions))
	out.BindingFailureConditions = *(*[]string)(unsafe.Pointer(&in.BindingFailureConditions))
	out.ShareID = (*types.UID)(unsafe.Pointer(in.ShareID))
	out.ConsumedCapacity = *(*map[resource.QualifiedName]apiresource.Quantity)(unsafe.Pointer(&in.ConsumedCapacity))
	return nil
}

// Convert_v1beta1_DeviceRequestAllocationResult_To_resource_DeviceRequestAllocationResult is an autogenerated conversion function.
func Convert_v1beta1_DeviceRequestAllocationResult_To_resource_DeviceRequestAllocationResult(in *resourcev1beta1.DeviceRequestAllocationResult, out *resource.DeviceRequestAllocationResult, s conversion.Scope) error {
	return autoConvert_v1beta1_DeviceRequestAllocationResult_To_resource_DeviceRequestAllocationResult(in, out, s)
}

func autoConvert_resource_DeviceRequestAllocationResult_To_v1beta1_DeviceRequestAllocationResult(in *resource.DeviceRequestAllocationResult, out *resourcev1beta1.DeviceRequestAllocationResult, s conversion.Scope) error {
	out.Request = in.Request
	out.Driver = in.Driver
	out.Pool = in.Pool
	out.Device = in.Device
	out.AdminAccess = (*bool)(unsafe.Pointer(in.AdminAccess))
	out.Tolerations = *(*[]resourcev1beta1.DeviceToleration)(unsafe.Pointer(&in.Tolerations))
	out.BindingConditions = *(*[]string)(unsafe.Pointer(&in.BindingConditions))
	out.BindingFailureConditions = *(*[]string)(unsafe.Pointer(&in.BindingFailureConditions))
	out.ShareID = (*types.UID)(unsafe.Pointer(in.ShareID))
	out.ConsumedCapacity = *(*map[resourcev1beta1.QualifiedName]apiresource.Quantity)(unsafe.Pointer(&in.ConsumedCapacity))
	return nil
}

// Convert_resource_DeviceRequestAllocationResult_To_v1beta1_DeviceRequestAllocationResult is an autogenerated conversion function.
func Convert_resource_DeviceRequestAllocationResult_To_v1beta1_DeviceRequestAllocationResult(in *resource.DeviceRequestAllocationResult, out *resourcev1beta1.DeviceRequestAllocationResult, s conversion.Scope) error {
	return autoConvert_resource_DeviceRequestAllocationResult_To_v1beta1_DeviceRequestAllocationResult(in, out, s)
}

func autoConvert_v1beta1_DeviceSelector_To_resource_DeviceSelector(in *resourcev1beta1.DeviceSelector, out *resource.DeviceSelector, s conversion.Scope) error {
	out.CEL = (*resource.CELDeviceSelector)(unsafe.Pointer(in.CEL))
	return nil
}

// Convert_v1beta1_DeviceSelector_To_resource_DeviceSelector is an autogenerated conversion function.
func Convert_v1beta1_DeviceSelector_To_resource_DeviceSelector(in *resourcev1beta1.DeviceSelector, out *resource.DeviceSelector, s conversion.Scope) error {
	return autoConvert_v1beta1_DeviceSelector_To_resource_DeviceSelector(in, out, s)
}

func autoConvert_resource_DeviceSelector_To_v1beta1_DeviceSelector(in *resource.DeviceSelector, out *resourcev1beta1.DeviceSelector, s conversion.Scope) error {
	out.CEL = (*resourcev1beta1.CELDeviceSelector)(unsafe.Pointer(in.CEL))
	return nil
}

// Convert_resource_DeviceSelector_To_v1beta1_DeviceSelector is an autogenerated conversion function.
func Convert_resource_DeviceSelector_To_v1beta1_DeviceSelector(in *resource.DeviceSelector, out *resourcev1beta1.DeviceSelector, s conversion.Scope) error {
	return autoConvert_resource_DeviceSelector_To_v1beta1_DeviceSelector(in, out, s)
}

func autoConvert_v1beta1_DeviceSubRequest_To_resource_DeviceSubRequest(in *resourcev1beta1.DeviceSubRequest, out *resource.DeviceSubRequest, s conversion.Scope) error {
	out.Name = in.Name
	out.DeviceClassName = in.DeviceClassName
	out.Selectors = *(*[]resource.DeviceSelector)(unsafe.Pointer(&in.Selectors))
	out.AllocationMode = resource.DeviceAllocationMode(in.AllocationMode)
	out.Count = in.Count
	out.Tolerations = *(*[]resource.DeviceToleration)(unsafe.Pointer(&in.Tolerations))
	out.Capacity = (*resource.CapacityRequirements)(unsafe.Pointer(in.Capacity))
	return nil
}

// Convert_v1beta1_DeviceSubRequest_To_resource_DeviceSubRequest is an autogenerated conversion function.
func Convert_v1beta1_DeviceSubRequest_To_resource_DeviceSubRequest(in *resourcev1beta1.DeviceSubRequest, out *resource.DeviceSubRequest, s conversion.Scope) error {
	return autoConvert_v1beta1_DeviceSubRequest_To_resource_DeviceSubRequest(in, out, s)
}

func autoConvert_resource_DeviceSubRequest_To_v1beta1_DeviceSubRequest(in *resource.DeviceSubRequest, out *resourcev1beta1.DeviceSubRequest, s conversion.Scope) error {
	out.Name = in.Name
	out.DeviceClassName = in.DeviceClassName
	out.Selectors = *(*[]resourcev1beta1.DeviceSelector)(unsafe.Pointer(&in.Selectors))
	out.AllocationMode = resourcev1beta1.DeviceAllocationMode(in.AllocationMode)
	out.Count = in.Count
	out.Tolerations = *(*[]resourcev1beta1.DeviceToleration)(unsafe.Pointer(&in.Tolerations))
	out.Capacity = (*resourcev1beta1.CapacityRequirements)(unsafe.Pointer(in.Capacity))
	return nil
}

// Convert_resource_DeviceSubRequest_To_v1beta1_DeviceSubRequest is an autogenerated conversion function.
func Convert_resource_DeviceSubRequest_To_v1beta1_DeviceSubRequest(in *resource.DeviceSubRequest, out *resourcev1beta1.DeviceSubRequest, s conversion.Scope) error {
	return autoConvert_resource_DeviceSubRequest_To_v1beta1_DeviceSubRequest(in, out, s)
}

func autoConvert_v1beta1_DeviceTaint_To_resource_DeviceTaint(in *resourcev1beta1.DeviceTaint, out *resource.DeviceTaint, s conversion.Scope) error {
	out.Key = in.Key
	out.Value = in.Value
	out.Effect = resource.DeviceTaintEffect(in.Effect)
	out.TimeAdded = (*v1.Time)(unsafe.Pointer(in.TimeAdded))
	return nil
}

// Convert_v1beta1_DeviceTaint_To_resource_DeviceTaint is an autogenerated conversion function.
func Convert_v1beta1_DeviceTaint_To_resource_DeviceTaint(in *resourcev1beta1.DeviceTaint, out *resource.DeviceTaint, s conversion.Scope) error {
	return autoConvert_v1beta1_DeviceTaint_To_resource_DeviceTaint(in, out, s)
}

func autoConvert_resource_DeviceTaint_To_v1beta1_DeviceTaint(in *resource.DeviceTaint, out *resourcev1beta1.DeviceTaint, s conversion.Scope) error {
	out.Key = in.Key
	out.Value = in.Value
	out.Effect = resourcev1beta1.DeviceTaintEffect(in.Effect)
	out.TimeAdded = (*v1.Time)(unsafe.Pointer(in.TimeAdded))
	return nil
}

// Convert_resource_DeviceTaint_To_v1beta1_DeviceTaint is an autogenerated conversion function.
func Convert_resource_DeviceTaint_To_v1beta1_DeviceTaint(in *resource.DeviceTaint, out *resourcev1beta1.DeviceTaint, s conversion.Scope) error {
	return autoConvert_resource_DeviceTaint_To_v1beta1_DeviceTaint(in, out, s)
}

func autoConvert_v1beta1_DeviceToleration_To_resource_DeviceToleration(in *resourcev1beta1.DeviceToleration, out *resource.DeviceToleration, s conversion.Scope) error {
	out.Key = in.Key
	out.Operator = resource.DeviceTolerationOperator(in.Operator)
	out.Value = in.Value
	out.Effect = resource.DeviceTaintEffect(in.Effect)
	out.TolerationSeconds = (*int64)(unsafe.Pointer(in.TolerationSeconds))
	return nil
}

// Convert_v1beta1_DeviceToleration_To_resource_DeviceToleration is an autogenerated conversion function.
func Convert_v1beta1_DeviceToleration_To_resource_DeviceToleration(in *resourcev1beta1.DeviceToleration, out *resource.DeviceToleration, s conversion.Scope) error {
	return autoConvert_v1beta1_DeviceToleration_To_resource_DeviceToleration(in, out, s)
}

func autoConvert_resource_DeviceToleration_To_v1beta1_DeviceToleration(in *resource.DeviceToleration, out *resourcev1beta1.DeviceToleration, s conversion.Scope) error {
	out.Key = in.Key
	out.Operator = resourcev1beta1.DeviceTolerationOperator(in.Operator)
	out.Value = in.Value
	out.Effect = resourcev1beta1.DeviceTaintEffect(in.Effect)
	out.TolerationSeconds = (*int64)(unsafe.Pointer(in.TolerationSeconds))
	return nil
}

// Convert_resource_DeviceToleration_To_v1beta1_DeviceToleration is an autogenerated conversion function.
func Convert_resource_DeviceToleration_To_v1beta1_DeviceToleration(in *resource.DeviceToleration, out *resourcev1beta1.DeviceToleration, s conversion.Scope) error {
	return autoConvert_resource_DeviceToleration_To_v1beta1_DeviceToleration(in, out, s)
}

func autoConvert_v1beta1_NetworkDeviceData_To_resource_NetworkDeviceData(in *resourcev1beta1.NetworkDeviceData, out *resource.NetworkDeviceData, s conversion.Scope) error {
	out.InterfaceName = in.InterfaceName
	out.IPs = *(*[]string)(unsafe.Pointer(&in.IPs))
	out.HardwareAddress = in.HardwareAddress
	return nil
}

// Convert_v1beta1_NetworkDeviceData_To_resource_NetworkDeviceData is an autogenerated conversion function.
func Convert_v1beta1_NetworkDeviceData_To_resource_NetworkDeviceData(in *resourcev1beta1.NetworkDeviceData, out *resource.NetworkDeviceData, s conversion.Scope) error {
	return autoConvert_v1beta1_NetworkDeviceData_To_resource_NetworkDeviceData(in, out, s)
}

func autoConvert_resource_NetworkDeviceData_To_v1beta1_NetworkDeviceData(in *resource.NetworkDeviceData, out *resourcev1beta1.NetworkDeviceData, s conversion.Scope) error {
	out.InterfaceName = in.InterfaceName
	out.IPs = *(*[]string)(unsafe.Pointer(&in.IPs))
	out.HardwareAddress = in.HardwareAddress
	return nil
}

// Convert_resource_NetworkDeviceData_To_v1beta1_NetworkDeviceData is an autogenerated conversion function.
func Convert_resource_NetworkDeviceData_To_v1beta1_NetworkDeviceData(in *resource.NetworkDeviceData, out *resourcev1beta1.NetworkDeviceData, s conversion.Scope) error {
	return autoConvert_resource_NetworkDeviceData_To_v1beta1_NetworkDeviceData(in, out, s)
}

func autoConvert_v1beta1_NodeAllocatableResourceMapping_To_resource_NodeAllocatableResourceMapping(in *resourcev1beta1.NodeAllocatableResourceMapping, out *resource.NodeAllocatableResourceMapping, s conversion.Scope) error {
	out.CapacityKey = (*resource.QualifiedName)(unsafe.Pointer(in.CapacityKey))
	out.AllocationMultiplier = (*apiresource.Quantity)(unsafe.Pointer(in.AllocationMultiplier))
	return nil
}

// Convert_v1beta1_NodeAllocatableResourceMapping_To_resource_NodeAllocatableResourceMapping is an autogenerated conversion function.
func Convert_v1beta1_NodeAllocatableResourceMapping_To_resource_NodeAllocatableResourceMapping(in *resourcev1beta1.NodeAllocatableResourceMapping, out *resource.NodeAllocatableResourceMapping, s conversion.Scope) error {
	return autoConvert_v1beta1_NodeAllocatableResourceMapping_To_resource_NodeAllocatableResourceMapping(in, out, s)
}

func autoConvert_resource_NodeAllocatableResourceMapping_To_v1beta1_NodeAllocatableResourceMapping(in *resource.NodeAllocatableResourceMapping, out *resourcev1beta1.NodeAllocatableResourceMapping, s conversion.Scope) error {
	out.CapacityKey = (*resourcev1beta1.QualifiedName)(unsafe.Pointer(in.CapacityKey))
	out.AllocationMultiplier = (*apiresource.Quantity)(unsafe.Pointer(in.AllocationMultiplier))
	return nil
}

// Convert_resource_NodeAllocatableResourceMapping_To_v1beta1_NodeAllocatableResourceMapping is an autogenerated conversion function.
func Convert_resource_NodeAllocatableResourceMapping_To_v1beta1_NodeAllocatableResourceMapping(in *resource.NodeAllocatableResourceMapping, out *resourcev1beta1.NodeAllocatableResourceMapping, s conversion.Scope) error {
	return autoConvert_resource_NodeAllocatableResourceMapping_To_v1beta1_NodeAllocatableResourceMapping(in, out, s)
}

func autoConvert_v1beta1_OpaqueDeviceConfiguration_To_resource_OpaqueDeviceConfiguration(in *resourcev1beta1.OpaqueDeviceConfiguration, out *resource.OpaqueDeviceConfiguration, s conversion.Scope) error {
	out.Driver = in.Driver
	out.Parameters = in.Parameters
	return nil
}

// Convert_v1beta1_OpaqueDeviceConfiguration_To_resource_OpaqueDeviceConfiguration is an autogenerated conversion function.
func Convert_v1beta1_OpaqueDeviceConfiguration_To_resource_OpaqueDeviceConfiguration(in *resourcev1beta1.OpaqueDeviceConfiguration, out *resource.OpaqueDeviceConfiguration, s conversion.Scope) error {
	return autoConvert_v1beta1_OpaqueDeviceConfiguration_To_resource_OpaqueDeviceConfiguration(in, out, s)
}

func autoConvert_resource_OpaqueDeviceConfiguration_To_v1beta1_OpaqueDeviceConfiguration(in *resource.OpaqueDeviceConfiguration, out *resourcev1beta1.OpaqueDeviceConfiguration, s conversion.Scope) error {
	out.Driver = in.Driver
	out.Parameters = in.Parameters
	return nil
}

// Convert_resource_OpaqueDeviceConfiguration_To_v1beta1_OpaqueDeviceConfiguration is an autogenerated conversion function.
func Convert_resource_OpaqueDeviceConfiguration_To_v1beta1_OpaqueDeviceConfiguration(in *resource.OpaqueDeviceConfiguration, out *resourcev1beta1.OpaqueDeviceConfiguration, s conversion.Scope) error {
	return autoConvert_resource_OpaqueDeviceConfiguration_To_v1beta1_OpaqueDeviceConfiguration(in, out, s)
}

func autoConvert_v1beta1_ResourceClaim_To_resource_ResourceClaim(in *resourcev1beta1.ResourceClaim, out *resource.ResourceClaim, s conversion.Scope) error {
	out.ObjectMeta = in.ObjectMeta
	if err := Convert_v1beta1_ResourceClaimSpec_To_resource_ResourceClaimSpec(&in.Spec, &out.Spec, s); err != nil {
		return err
	}
	if err := Convert_v1beta1_ResourceClaimStatus_To_resource_ResourceClaimStatus(&in.Status, &out.Status, s); err != nil {
		return err
	}
	return nil
}

// Convert_v1beta1_ResourceClaim_To_resource_ResourceClaim is an autogenerated conversion function.
func Convert_v1beta1_ResourceClaim_To_resource_ResourceClaim(in *resourcev1beta1.ResourceClaim, out *resource.ResourceClaim, s conversion.Scope) error {
	return autoConvert_v1beta1_ResourceClaim_To_resource_ResourceClaim(in, out, s)
}

func autoConvert_resource_ResourceClaim_To_v1beta1_ResourceClaim(in *resource.ResourceClaim, out *resourcev1beta1.ResourceClaim, s conversion.Scope) error {
	out.ObjectMeta = in.ObjectMeta
	if err := Convert_resource_ResourceClaimSpec_To_v1beta1_ResourceClaimSpec(&in.Spec, &out.Spec, s); err != nil {
		return err
	}
	if err := Convert_resource_ResourceClaimStatus_To_v1beta1_ResourceClaimStatus(&in.Status, &out.Status, s); err != nil {
		return err
	}
	return nil
}

// Convert_resource_ResourceClaim_To_v1beta1_ResourceClaim is an autogenerated conversion function.
func Convert_resource_ResourceClaim_To_v1beta1_ResourceClaim(in *resource.ResourceClaim, out *resourcev1beta1.ResourceClaim, s conversion.Scope) error {
	return autoConvert_resource_ResourceClaim_To_v1beta1_ResourceClaim(in, out, s)
}

func autoConvert_v1beta1_ResourceClaimConsumerReference_To_resource_ResourceClaimConsumerReference(in *resourcev1beta1.ResourceClaimConsumerReference, out *resource.ResourceClaimConsumerReference, s conversion.Scope) error {
	out.APIGroup = in.APIGroup
	out.Resource = in.Resource
	out.Name = in.Name
	out.UID = types.UID(in.UID)
	return nil
}

// Convert_v1beta1_ResourceClaimConsumerReference_To_resource_ResourceClaimConsumerReference is an autogenerated conversion function.
func Convert_v1beta1_ResourceClaimConsumerReference_To_resource_ResourceClaimConsumerReference(in *resourcev1beta1.ResourceClaimConsumerReference, out *resource.ResourceClaimConsumerReference, s conversion.Scope) error {
	return autoConvert_v1beta1_ResourceClaimConsumerReference_To_resource_ResourceClaimConsumerReference(in, out, s)
}

func autoConvert_resource_ResourceClaimConsumerReference_To_v1beta1_ResourceClaimConsumerReference(in *resource.ResourceClaimConsumerReference, out *resourcev1beta1.ResourceClaimConsumerReference, s conversion.Scope) error {
	out.APIGroup = in.APIGroup
	out.Resource = in.Resource
	out.Name = in.Name
	out.UID = types.UID(in.UID)
	return nil
}

// Convert_resource_ResourceClaimConsumerReference_To_v1beta1_ResourceClaimConsumerReference is an autogenerated conversion function.
func Convert_resource_ResourceClaimConsumerReference_To_v1beta1_ResourceClaimConsumerReference(in *resource.ResourceClaimConsumerReference, out *resourcev1beta1.ResourceClaimConsumerReference, s conversion.Scope) error {
	return autoConvert_resource_ResourceClaimConsumerReference_To_v1beta1_ResourceClaimConsumerReference(in, out, s)
}

func autoConvert_v1beta1_ResourceClaimList_To_resource_ResourceClaimList(in *resourcev1beta1.ResourceClaimList, out *resource.ResourceClaimList, s conversion.Scope) error {
	out.ListMeta = in.ListMeta
	if in.Items != nil {
		in, out := &in.Items, &out.Items
		*out = make([]resource.ResourceClaim, len(*in))
		for i := range *in {
			if err := Convert_v1beta1_ResourceClaim_To_resource_ResourceClaim(&(*in)[i], &(*out)[i], s); err != nil {
				return err
			}
		}
	} else {
		out.Items = nil
	}
	return nil
}

// Convert_v1beta1_ResourceClaimList_To_resource_ResourceClaimList is an autogenerated conversion function.
func Convert_v1beta1_ResourceClaimList_To_resource_ResourceClaimList(in *resourcev1beta1.ResourceClaimList, out *resource.ResourceClaimList, s conversion.Scope) error {
	return autoConvert_v1beta1_ResourceClaimList_To_resource_ResourceClaimList(in, out, s)
}

func autoConvert_resource_ResourceClaimList_To_v1beta1_ResourceClaimList(in *resource.ResourceClaimList, out *resourcev1beta1.ResourceClaimList, s conversion.Scope) error {
	out.ListMeta = in.ListMeta
	if in.Items != nil {
		in, out := &in.Items, &out.Items
		*out = make([]resourcev1beta1.ResourceClaim, len(*in))
		for i := range *in {
			if err := Convert_resource_ResourceClaim_To_v1beta1_ResourceClaim(&(*in)[i], &(*out)[i], s); err != nil {
				return err
			}
		}
	} else {
		out.Items = nil
	}
	return nil
}

// Convert_resource_ResourceClaimList_To_v1beta1_ResourceClaimList is an autogenerated conversion function.
func Convert_resource_ResourceClaimList_To_v1beta1_ResourceClaimList(in *resource.ResourceClaimList, out *resourcev1beta1.ResourceClaimList, s conversion.Scope) error {
	return autoConvert_resource_ResourceClaimList_To_v1beta1_ResourceClaimList(in, out, s)
}

func autoConvert_v1beta1_ResourceClaimSpec_To_resource_ResourceClaimSpec(in *resourcev1beta1.ResourceClaimSpec, out *resource.ResourceClaimSpec, s conversion.Scope) error {
	if err := Convert_v1beta1_DeviceClaim_To_resource_DeviceClaim(&in.Devices, &out.Devices, s); err != nil {
		return err
	}
	return nil
}

// Convert_v1beta1_ResourceClaimSpec_To_resource_ResourceClaimSpec is an autogenerated conversion function.
func Convert_v1beta1_ResourceClaimSpec_To_resource_ResourceClaimSpec(in *resourcev1beta1.ResourceClaimSpec, out *resource.ResourceClaimSpec, s conversion.Scope) error {
	return autoConvert_v1beta1_ResourceClaimSpec_To_resource_ResourceClaimSpec(in, out, s)
}

func autoConvert_resource_ResourceClaimSpec_To_v1beta1_ResourceClaimSpec(in *resource.ResourceClaimSpec, out *resourcev1beta1.ResourceClaimSpec, s conversion.Scope) error {
	if err := Convert_resource_DeviceClaim_To_v1beta1_DeviceClaim(&in.Devices, &out.Devices, s); err != nil {
		return err
	}
	return nil
}

// Convert_resource_ResourceClaimSpec_To_v1beta1_ResourceClaimSpec is an autogenerated conversion function.
func Convert_resource_ResourceClaimSpec_To_v1beta1_ResourceClaimSpec(in *resource.ResourceClaimSpec, out *resourcev1beta1.ResourceClaimSpec, s conversion.Scope) error {
	return autoConvert_resource_ResourceClaimSpec_To_v1beta1_ResourceClaimSpec(in, out, s)
}

func autoConvert_v1beta1_ResourceClaimStatus_To_resource_ResourceClaimStatus(in *resourcev1beta1.ResourceClaimStatus, out *resource.ResourceClaimStatus, s conversion.Scope) error {
	out.Allocation = (*resource.AllocationResult)(unsafe.Pointer(in.Allocation))
	out.ReservedFor = *(*[]resource.ResourceClaimConsumerReference)(unsafe.Pointer(&in.ReservedFor))
	out.Devices = *(*[]resource.AllocatedDeviceStatus)(unsafe.Pointer(&in.Devices))
	return nil
}

// Convert_v1beta1_ResourceClaimStatus_To_resource_ResourceClaimStatus is an autogenerated conversion function.
func Convert_v1beta1_ResourceClaimStatus_To_resource_ResourceClaimStatus(in *resourcev1beta1.ResourceClaimStatus, out *resource.ResourceClaimStatus, s conversion.Scope) error {
	return autoConvert_v1beta1_ResourceClaimStatus_To_resource_ResourceClaimStatus(in, out, s)
}

func autoConvert_resource_ResourceClaimStatus_To_v1beta1_ResourceClaimStatus(in *resource.ResourceClaimStatus, out *resourcev1beta1.ResourceClaimStatus, s conversion.Scope) error {
	out.Allocation = (*resourcev1beta1.AllocationResult)(unsafe.Pointer(in.Allocation))
	out.ReservedFor = *(*[]resourcev1beta1.ResourceClaimConsumerReference)(unsafe.Pointer(&in.ReservedFor))
	out.Devices = *(*[]resourcev1beta1.AllocatedDeviceStatus)(unsafe.Pointer(&in.Devices))
	return nil
}

// Convert_resource_ResourceClaimStatus_To_v1beta1_ResourceClaimStatus is an autogenerated conversion function.
func Convert_resource_ResourceClaimStatus_To_v1beta1_ResourceClaimStatus(in *resource.ResourceClaimStatus, out *resourcev1beta1.ResourceClaimStatus, s conversion.Scope) error {
	return autoConvert_resource_ResourceClaimStatus_To_v1beta1_ResourceClaimStatus(in, out, s)
}

func autoConvert_v1beta1_ResourceClaimTemplate_To_resource_ResourceClaimTemplate(in *resourcev1beta1.ResourceClaimTemplate, out *resource.ResourceClaimTemplate, s conversion.Scope) error {
	out.ObjectMeta = in.ObjectMeta
	if err := Convert_v1beta1_ResourceClaimTemplateSpec_To_resource_ResourceClaimTemplateSpec(&in.Spec, &out.Spec, s); err != nil {
		return err
	}
	return nil
}

// Convert_v1beta1_ResourceClaimTemplate_To_resource_ResourceClaimTemplate is an autogenerated conversion function.
func Convert_v1beta1_ResourceClaimTemplate_To_resource_ResourceClaimTemplate(in *resourcev1beta1.ResourceClaimTemplate, out *resource.ResourceClaimTemplate, s conversion.Scope) error {
	return autoConvert_v1beta1_ResourceClaimTemplate_To_resource_ResourceClaimTemplate(in, out, s)
}

func autoConvert_resource_ResourceClaimTemplate_To_v1beta1_ResourceClaimTemplate(in *resource.ResourceClaimTemplate, out *resourcev1beta1.ResourceClaimTemplate, s conversion.Scope) error {
	out.ObjectMeta = in.ObjectMeta
	if err := Convert_resource_ResourceClaimTemplateSpec_To_v1beta1_ResourceClaimTemplateSpec(&in.Spec, &out.Spec, s); err != nil {
		return err
	}
	return nil
}

// Convert_resource_ResourceClaimTemplate_To_v1beta1_ResourceClaimTemplate is an autogenerated conversion function.
func Convert_resource_ResourceClaimTemplate_To_v1beta1_ResourceClaimTemplate(in *resource.ResourceClaimTemplate, out *resourcev1beta1.ResourceClaimTemplate, s conversion.Scope) error {
	return autoConvert_resource_ResourceClaimTemplate_To_v1beta1_ResourceClaimTemplate(in, out, s)
}

func autoConvert_v1beta1_ResourceClaimTemplateList_To_resource_ResourceClaimTemplateList(in *resourcev1beta1.ResourceClaimTemplateList, out *resource.ResourceClaimTemplateList, s conversion.Scope) error {
	out.ListMeta = in.ListMeta
	if in.Items != nil {
		in, out := &in.Items, &out.Items
		*out = make([]resource.ResourceClaimTemplate, len(*in))
		for i := range *in {
			if err := Convert_v1beta1_ResourceClaimTemplate_To_resource_ResourceClaimTemplate(&(*in)[i], &(*out)[i], s); err != nil {
				return err
			}
		}
	} else {
		out.Items = nil
	}
	return nil
}

// Convert_v1beta1_ResourceClaimTemplateList_To_resource_ResourceClaimTemplateList is an autogenerated conversion function.
func Convert_v1beta1_ResourceClaimTemplateList_To_resource_ResourceClaimTemplateList(in *resourcev1beta1.ResourceClaimTemplateList, out *resource.ResourceClaimTemplateList, s conversion.Scope) error {
	return autoConvert_v1beta1_ResourceClaimTemplateList_To_resource_ResourceClaimTemplateList(in, out, s)
}

func autoConvert_resource_ResourceClaimTemplateList_To_v1beta1_ResourceClaimTemplateList(in *resource.ResourceClaimTemplateList, out *resourcev1beta1.ResourceClaimTemplateList, s conversion.Scope) error {
	out.ListMeta = in.ListMeta
	if in.Items != nil {
		in, out := &in.Items, &out.Items
		*out = make([]resourcev1beta1.ResourceClaimTemplate, len(*in))
		for i := range *in {
			if err := Convert_resource_ResourceClaimTemplate_To_v1beta1_ResourceClaimTemplate(&(*in)[i], &(*out)[i], s); err != nil {
				return err
			}
		}
	} else {
		out.Items = nil
	}
	return nil
}

// Convert_resource_ResourceClaimTemplateList_To_v1beta1_ResourceClaimTemplateList is an autogenerated conversion function.
func Convert_resource_ResourceClaimTemplateList_To_v1beta1_ResourceClaimTemplateList(in *resource.ResourceClaimTemplateList, out *resourcev1beta1.ResourceClaimTemplateList, s conversion.Scope) error {
	return autoConvert_resource_ResourceClaimTemplateList_To_v1beta1_ResourceClaimTemplateList(in, out, s)
}

func autoConvert_v1beta1_ResourceClaimTemplateSpec_To_resource_ResourceClaimTemplateSpec(in *resourcev1beta1.ResourceClaimTemplateSpec, out *resource.ResourceClaimTemplateSpec, s conversion.Scope) error {
	out.ObjectMeta = in.ObjectMeta
	if err := Convert_v1beta1_ResourceClaimSpec_To_resource_ResourceClaimSpec(&in.Spec, &out.Spec, s); err != nil {
		return err
	}
	return nil
}

// Convert_v1beta1_ResourceClaimTemplateSpec_To_resource_ResourceClaimTemplateSpec is an autogenerated conversion function.
func Convert_v1beta1_ResourceClaimTemplateSpec_To_resource_ResourceClaimTemplateSpec(in *resourcev1beta1.ResourceClaimTemplateSpec, out *resource.ResourceClaimTemplateSpec, s conversion.Scope) error {
	return autoConvert_v1beta1_ResourceClaimTemplateSpec_To_resource_ResourceClaimTemplateSpec(in, out, s)
}

func autoConvert_resource_ResourceClaimTemplateSpec_To_v1beta1_ResourceClaimTemplateSpec(in *resource.ResourceClaimTemplateSpec, out *resourcev1beta1.ResourceClaimTemplateSpec, s conversion.Scope) error {
	out.ObjectMeta = in.ObjectMeta
	if err := Convert_resource_ResourceClaimSpec_To_v1beta1_ResourceClaimSpec(&in.Spec, &out.Spec, s); err != nil {
		return err
	}
	return nil
}

// Convert_resource_ResourceClaimTemplateSpec_To_v1beta1_ResourceClaimTemplateSpec is an autogenerated conversion function.
func Convert_resource_ResourceClaimTemplateSpec_To_v1beta1_ResourceClaimTemplateSpec(in *resource.ResourceClaimTemplateSpec, out *resourcev1beta1.ResourceClaimTemplateSpec, s conversion.Scope) error {
	return autoConvert_resource_ResourceClaimTemplateSpec_To_v1beta1_ResourceClaimTemplateSpec(in, out, s)
}

func autoConvert_v1beta1_ResourcePool_To_resource_ResourcePool(in *resourcev1beta1.ResourcePool, out *resource.ResourcePool, s conversion.Scope) error {
	out.Name = in.Name
	out.Generation = in.Generation
	out.ResourceSliceCount = in.ResourceSliceCount
	return nil
}

// Convert_v1beta1_ResourcePool_To_resource_ResourcePool is an autogenerated conversion function.
func Convert_v1beta1_ResourcePool_To_resource_ResourcePool(in *resourcev1beta1.ResourcePool, out *resource.ResourcePool, s conversion.Scope) error {
	return autoConvert_v1beta1_ResourcePool_To_resource_ResourcePool(in, out, s)
}

func autoConvert_resource_ResourcePool_To_v1beta1_ResourcePool(in *resource.ResourcePool, out *resourcev1beta1.ResourcePool, s conversion.Scope) error {
	out.Name = in.Name
	out.Generation = in.Generation
	out.ResourceSliceCount = in.ResourceSliceCount
	return nil
}

// Convert_resource_ResourcePool_To_v1beta1_ResourcePool is an autogenerated conversion function.
func Convert_resource_ResourcePool_To_v1beta1_ResourcePool(in *resource.ResourcePool, out *resourcev1beta1.ResourcePool, s conversion.Scope) error {
	return autoConvert_resource_ResourcePool_To_v1beta1_ResourcePool(in, out, s)
}

func autoConvert_v1beta1_ResourceSlice_To_resource_ResourceSlice(in *resourcev1beta1.ResourceSlice, out *resource.ResourceSlice, s conversion.Scope) error {
	out.ObjectMeta = in.ObjectMeta
	if err := Convert_v1beta1_ResourceSliceSpec_To_resource_ResourceSliceSpec(&in.Spec, &out.Spec, s); err != nil {
		return err
	}
	return nil
}

// Convert_v1beta1_ResourceSlice_To_resource_ResourceSlice is an autogenerated conversion function.
func Convert_v1beta1_ResourceSlice_To_resource_ResourceSlice(in *resourcev1beta1.ResourceSlice, out *resource.ResourceSlice, s conversion.Scope) error {
	return autoConvert_v1beta1_ResourceSlice_To_resource_ResourceSlice(in, out, s)
}

func autoConvert_resource_ResourceSlice_To_v1beta1_ResourceSlice(in *resource.ResourceSlice, out *resourcev1beta1.ResourceSlice, s conversion.Scope) error {
	out.ObjectMeta = in.ObjectMeta
	if err := Convert_resource_ResourceSliceSpec_To_v1beta1_ResourceSliceSpec(&in.Spec, &out.Spec, s); err != nil {
		return err
	}
	return nil
}

// Convert_resource_ResourceSlice_To_v1beta1_ResourceSlice is an autogenerated conversion function.
func Convert_resource_ResourceSlice_To_v1beta1_ResourceSlice(in *resource.ResourceSlice, out *resourcev1beta1.ResourceSlice, s conversion.Scope) error {
	return autoConvert_resource_ResourceSlice_To_v1beta1_ResourceSlice(in, out, s)
}

func autoConvert_v1beta1_ResourceSliceList_To_resource_ResourceSliceList(in *resourcev1beta1.ResourceSliceList, out *resource.ResourceSliceList, s conversion.Scope) error {
	out.ListMeta = in.ListMeta
	if in.Items != nil {
		in, out := &in.Items, &out.Items
		*out = make([]resource.ResourceSlice, len(*in))
		for i := range *in {
			if err := Convert_v1beta1_ResourceSlice_To_resource_ResourceSlice(&(*in)[i], &(*out)[i], s); err != nil {
				return err
			}
		}
	} else {
		out.Items = nil
	}
	return nil
}

// Convert_v1beta1_ResourceSliceList_To_resource_ResourceSliceList is an autogenerated conversion function.
func Convert_v1beta1_ResourceSliceList_To_resource_ResourceSliceList(in *resourcev1beta1.ResourceSliceList, out *resource.ResourceSliceList, s conversion.Scope) error {
	return autoConvert_v1beta1_ResourceSliceList_To_resource_ResourceSliceList(in, out, s)
}

func autoConvert_resource_ResourceSliceList_To_v1beta1_ResourceSliceList(in *resource.ResourceSliceList, out *resourcev1beta1.ResourceSliceList, s conversion.Scope) error {
	out.ListMeta = in.ListMeta
	if in.Items != nil {
		in, out := &in.Items, &out.Items
		*out = make([]resourcev1beta1.ResourceSlice, len(*in))
		for i := range *in {
			if err := Convert_resource_ResourceSlice_To_v1beta1_ResourceSlice(&(*in)[i], &(*out)[i], s); err != nil {
				return err
			}
		}
	} else {
		out.Items = nil
	}
	return nil
}

// Convert_resource_ResourceSliceList_To_v1beta1_ResourceSliceList is an autogenerated conversion function.
func Convert_resource_ResourceSliceList_To_v1beta1_ResourceSliceList(in *resource.ResourceSliceList, out *resourcev1beta1.ResourceSliceList, s conversion.Scope) error {
	return autoConvert_resource_ResourceSliceList_To_v1beta1_ResourceSliceList(in, out, s)
}

func autoConvert_v1beta1_ResourceSliceSpec_To_resource_ResourceSliceSpec(in *resourcev1beta1.ResourceSliceSpec, out *resource.ResourceSliceSpec, s conversion.Scope) error {
	out.Driver = in.Driver
	if err := Convert_v1beta1_ResourcePool_To_resource_ResourcePool(&in.Pool, &out.Pool, s); err != nil {
		return err
	}
	if err := v1.Convert_string_To_Pointer_string(&in.NodeName, &out.NodeName, s); err != nil {
		return err
	}
	out.NodeSelector = (*core.NodeSelector)(unsafe.Pointer(in.NodeSelector))
	if err := v1.Convert_bool_To_Pointer_bool(&in.AllNodes, &out.AllNodes, s); err != nil {
		return err
	}
	if in.Devices != nil {
		in, out := &in.Devices, &out.Devices
		*out = make([]resource.Device, len(*in))
		for i := range *in {
			if err := Convert_v1beta1_Device_To_resource_Device(&(*in)[i], &(*out)[i], s); err != nil {
				return err
			}
		}
	} else {
		out.Devices = nil
	}
	out.PerDeviceNodeSelection = (*bool)(unsafe.Pointer(in.PerDeviceNodeSelection))
	out.SharedCounters = *(*[]resource.CounterSet)(unsafe.Pointer(&in.SharedCounters))
	return nil
}

func autoConvert_resource_ResourceSliceSpec_To_v1beta1_ResourceSliceSpec(in *resource.ResourceSliceSpec, out *resourcev1beta1.ResourceSliceSpec, s conversion.Scope) error {
	out.Driver = in.Driver
	if err := Convert_resource_ResourcePool_To_v1beta1_ResourcePool(&in.Pool, &out.Pool, s); err != nil {
		return err
	}
	if err := v1.Convert_Pointer_string_To_string(&in.NodeName, &out.NodeName, s); err != nil {
		return err
	}
	out.NodeSelector = (*corev1.NodeSelector)(unsafe.Pointer(in.NodeSelector))
	if err := v1.Convert_Pointer_bool_To_bool(&in.AllNodes, &out.AllNodes, s); err != nil {
		return err
	}
	if in.Devices != nil {
		in, out := &in.Devices, &out.Devices
		*out = make([]resourcev1beta1.Device, len(*in))
		for i := range *in {
			if err := Convert_resource_Device_To_v1beta1_Device(&(*in)[i], &(*out)[i], s); err != nil {
				return err
			}
		}
	} else {
		out.Devices = nil
	}
	out.PerDeviceNodeSelection = (*bool)(unsafe.Pointer(in.PerDeviceNodeSelection))
	out.SharedCounters = *(*[]resourcev1beta1.CounterSet)(unsafe.Pointer(&in.SharedCounters))
	return nil
}
