/*
Copyright 2023 The Kubernetes Authors.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
*/

package common

// QuotaID is generic quota identifier.
// Data type based on quotactl(2).
type QuotaID int32

const (
	// UnknownQuotaID -- cannot determine whether a quota is in force
	UnknownQuotaID QuotaID = -1
	// BadQuotaID -- Invalid quota
	BadQuotaID QuotaID = 0
)
