//go:build !ignore_autogenerated
// +build !ignore_autogenerated

/*
Copyright The Kubernetes Authors.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
*/

// Code generated by conversion-gen. DO NOT EDIT.

package v1alpha2

import (
	unsafe "unsafe"

	schedulingv1alpha2 "k8s.io/api/scheduling/v1alpha2"
	v1 "k8s.io/apimachinery/pkg/apis/meta/v1"
	conversion "k8s.io/apimachinery/pkg/conversion"
	runtime "k8s.io/apimachinery/pkg/runtime"
	scheduling "k8s.io/kubernetes/pkg/apis/scheduling"
)

func init() {
	localSchemeBuilder.Register(RegisterConversions)
}

// RegisterConversions adds conversion functions to the given scheme.
// Public to allow building arbitrary schemes.
func RegisterConversions(s *runtime.Scheme) error {
	if err := s.AddGeneratedConversionFunc((*schedulingv1alpha2.BasicSchedulingPolicy)(nil), (*scheduling.BasicSchedulingPolicy)(nil), func(a, b interface{}, scope conversion.Scope) error {
		return Convert_v1alpha2_BasicSchedulingPolicy_To_scheduling_BasicSchedulingPolicy(a.(*schedulingv1alpha2.BasicSchedulingPolicy), b.(*scheduling.BasicSchedulingPolicy), scope)
	}); err != nil {
		return err
	}
	if err := s.AddGeneratedConversionFunc((*scheduling.BasicSchedulingPolicy)(nil), (*schedulingv1alpha2.BasicSchedulingPolicy)(nil), func(a, b interface{}, scope conversion.Scope) error {
		return Convert_scheduling_BasicSchedulingPolicy_To_v1alpha2_BasicSchedulingPolicy(a.(*scheduling.BasicSchedulingPolicy), b.(*schedulingv1alpha2.BasicSchedulingPolicy), scope)
	}); err != nil {
		return err
	}
	if err := s.AddGeneratedConversionFunc((*schedulingv1alpha2.GangSchedulingPolicy)(nil), (*scheduling.GangSchedulingPolicy)(nil), func(a, b interface{}, scope conversion.Scope) error {
		return Convert_v1alpha2_GangSchedulingPolicy_To_scheduling_GangSchedulingPolicy(a.(*schedulingv1alpha2.GangSchedulingPolicy), b.(*scheduling.GangSchedulingPolicy), scope)
	}); err != nil {
		return err
	}
	if err := s.AddGeneratedConversionFunc((*scheduling.GangSchedulingPolicy)(nil), (*schedulingv1alpha2.GangSchedulingPolicy)(nil), func(a, b interface{}, scope conversion.Scope) error {
		return Convert_scheduling_GangSchedulingPolicy_To_v1alpha2_GangSchedulingPolicy(a.(*scheduling.GangSchedulingPolicy), b.(*schedulingv1alpha2.GangSchedulingPolicy), scope)
	}); err != nil {
		return err
	}
	if err := s.AddGeneratedConversionFunc((*schedulingv1alpha2.PodGroup)(nil), (*scheduling.PodGroup)(nil), func(a, b interface{}, scope conversion.Scope) error {
		return Convert_v1alpha2_PodGroup_To_scheduling_PodGroup(a.(*schedulingv1alpha2.PodGroup), b.(*scheduling.PodGroup), scope)
	}); err != nil {
		return err
	}
	if err := s.AddGeneratedConversionFunc((*scheduling.PodGroup)(nil), (*schedulingv1alpha2.PodGroup)(nil), func(a, b interface{}, scope conversion.Scope) error {
		return Convert_scheduling_PodGroup_To_v1alpha2_PodGroup(a.(*scheduling.PodGroup), b.(*schedulingv1alpha2.PodGroup), scope)
	}); err != nil {
		return err
	}
	if err := s.AddGeneratedConversionFunc((*schedulingv1alpha2.PodGroupList)(nil), (*scheduling.PodGroupList)(nil), func(a, b interface{}, scope conversion.Scope) error {
		return Convert_v1alpha2_PodGroupList_To_scheduling_PodGroupList(a.(*schedulingv1alpha2.PodGroupList), b.(*scheduling.PodGroupList), scope)
	}); err != nil {
		return err
	}
	if err := s.AddGeneratedConversionFunc((*scheduling.PodGroupList)(nil), (*schedulingv1alpha2.PodGroupList)(nil), func(a, b interface{}, scope conversion.Scope) error {
		return Convert_scheduling_PodGroupList_To_v1alpha2_PodGroupList(a.(*scheduling.PodGroupList), b.(*schedulingv1alpha2.PodGroupList), scope)
	}); err != nil {
		return err
	}
	if err := s.AddGeneratedConversionFunc((*schedulingv1alpha2.PodGroupResourceClaim)(nil), (*scheduling.PodGroupResourceClaim)(nil), func(a, b interface{}, scope conversion.Scope) error {
		return Convert_v1alpha2_PodGroupResourceClaim_To_scheduling_PodGroupResourceClaim(a.(*schedulingv1alpha2.PodGroupResourceClaim), b.(*scheduling.PodGroupResourceClaim), scope)
	}); err != nil {
		return err
	}
	if err := s.AddGeneratedConversionFunc((*scheduling.PodGroupResourceClaim)(nil), (*schedulingv1alpha2.PodGroupResourceClaim)(nil), func(a, b interface{}, scope conversion.Scope) error {
		return Convert_scheduling_PodGroupResourceClaim_To_v1alpha2_PodGroupResourceClaim(a.(*scheduling.PodGroupResourceClaim), b.(*schedulingv1alpha2.PodGroupResourceClaim), scope)
	}); err != nil {
		return err
	}
	if err := s.AddGeneratedConversionFunc((*schedulingv1alpha2.PodGroupResourceClaimStatus)(nil), (*scheduling.PodGroupResourceClaimStatus)(nil), func(a, b interface{}, scope conversion.Scope) error {
		return Convert_v1alpha2_PodGroupResourceClaimStatus_To_scheduling_PodGroupResourceClaimStatus(a.(*schedulingv1alpha2.PodGroupResourceClaimStatus), b.(*scheduling.PodGroupResourceClaimStatus), scope)
	}); err != nil {
		return err
	}
	if err := s.AddGeneratedConversionFunc((*scheduling.PodGroupResourceClaimStatus)(nil), (*schedulingv1alpha2.PodGroupResourceClaimStatus)(nil), func(a, b interface{}, scope conversion.Scope) error {
		return Convert_scheduling_PodGroupResourceClaimStatus_To_v1alpha2_PodGroupResourceClaimStatus(a.(*scheduling.PodGroupResourceClaimStatus), b.(*schedulingv1alpha2.PodGroupResourceClaimStatus), scope)
	}); err != nil {
		return err
	}
	if err := s.AddGeneratedConversionFunc((*schedulingv1alpha2.PodGroupSchedulingConstraints)(nil), (*scheduling.PodGroupSchedulingConstraints)(nil), func(a, b interface{}, scope conversion.Scope) error {
		return Convert_v1alpha2_PodGroupSchedulingConstraints_To_scheduling_PodGroupSchedulingConstraints(a.(*schedulingv1alpha2.PodGroupSchedulingConstraints), b.(*scheduling.PodGroupSchedulingConstraints), scope)
	}); err != nil {
		return err
	}
	if err := s.AddGeneratedConversionFunc((*scheduling.PodGroupSchedulingConstraints)(nil), (*schedulingv1alpha2.PodGroupSchedulingConstraints)(nil), func(a, b interface{}, scope conversion.Scope) error {
		return Convert_scheduling_PodGroupSchedulingConstraints_To_v1alpha2_PodGroupSchedulingConstraints(a.(*scheduling.PodGroupSchedulingConstraints), b.(*schedulingv1alpha2.PodGroupSchedulingConstraints), scope)
	}); err != nil {
		return err
	}
	if err := s.AddGeneratedConversionFunc((*schedulingv1alpha2.PodGroupSchedulingPolicy)(nil), (*scheduling.PodGroupSchedulingPolicy)(nil), func(a, b interface{}, scope conversion.Scope) error {
		return Convert_v1alpha2_PodGroupSchedulingPolicy_To_scheduling_PodGroupSchedulingPolicy(a.(*schedulingv1alpha2.PodGroupSchedulingPolicy), b.(*scheduling.PodGroupSchedulingPolicy), scope)
	}); err != nil {
		return err
	}
	if err := s.AddGeneratedConversionFunc((*scheduling.PodGroupSchedulingPolicy)(nil), (*schedulingv1alpha2.PodGroupSchedulingPolicy)(nil), func(a, b interface{}, scope conversion.Scope) error {
		return Convert_scheduling_PodGroupSchedulingPolicy_To_v1alpha2_PodGroupSchedulingPolicy(a.(*scheduling.PodGroupSchedulingPolicy), b.(*schedulingv1alpha2.PodGroupSchedulingPolicy), scope)
	}); err != nil {
		return err
	}
	if err := s.AddGeneratedConversionFunc((*schedulingv1alpha2.PodGroupSpec)(nil), (*scheduling.PodGroupSpec)(nil), func(a, b interface{}, scope conversion.Scope) error {
		return Convert_v1alpha2_PodGroupSpec_To_scheduling_PodGroupSpec(a.(*schedulingv1alpha2.PodGroupSpec), b.(*scheduling.PodGroupSpec), scope)
	}); err != nil {
		return err
	}
	if err := s.AddGeneratedConversionFunc((*scheduling.PodGroupSpec)(nil), (*schedulingv1alpha2.PodGroupSpec)(nil), func(a, b interface{}, scope conversion.Scope) error {
		return Convert_scheduling_PodGroupSpec_To_v1alpha2_PodGroupSpec(a.(*scheduling.PodGroupSpec), b.(*schedulingv1alpha2.PodGroupSpec), scope)
	}); err != nil {
		return err
	}
	if err := s.AddGeneratedConversionFunc((*schedulingv1alpha2.PodGroupStatus)(nil), (*scheduling.PodGroupStatus)(nil), func(a, b interface{}, scope conversion.Scope) error {
		return Convert_v1alpha2_PodGroupStatus_To_scheduling_PodGroupStatus(a.(*schedulingv1alpha2.PodGroupStatus), b.(*scheduling.PodGroupStatus), scope)
	}); err != nil {
		return err
	}
	if err := s.AddGeneratedConversionFunc((*scheduling.PodGroupStatus)(nil), (*schedulingv1alpha2.PodGroupStatus)(nil), func(a, b interface{}, scope conversion.Scope) error {
		return Convert_scheduling_PodGroupStatus_To_v1alpha2_PodGroupStatus(a.(*scheduling.PodGroupStatus), b.(*schedulingv1alpha2.PodGroupStatus), scope)
	}); err != nil {
		return err
	}
	if err := s.AddGeneratedConversionFunc((*schedulingv1alpha2.PodGroupTemplate)(nil), (*scheduling.PodGroupTemplate)(nil), func(a, b interface{}, scope conversion.Scope) error {
		return Convert_v1alpha2_PodGroupTemplate_To_scheduling_PodGroupTemplate(a.(*schedulingv1alpha2.PodGroupTemplate), b.(*scheduling.PodGroupTemplate), scope)
	}); err != nil {
		return err
	}
	if err := s.AddGeneratedConversionFunc((*scheduling.PodGroupTemplate)(nil), (*schedulingv1alpha2.PodGroupTemplate)(nil), func(a, b interface{}, scope conversion.Scope) error {
		return Convert_scheduling_PodGroupTemplate_To_v1alpha2_PodGroupTemplate(a.(*scheduling.PodGroupTemplate), b.(*schedulingv1alpha2.PodGroupTemplate), scope)
	}); err != nil {
		return err
	}
	if err := s.AddGeneratedConversionFunc((*schedulingv1alpha2.PodGroupTemplateReference)(nil), (*scheduling.PodGroupTemplateReference)(nil), func(a, b interface{}, scope conversion.Scope) error {
		return Convert_v1alpha2_PodGroupTemplateReference_To_scheduling_PodGroupTemplateReference(a.(*schedulingv1alpha2.PodGroupTemplateReference), b.(*scheduling.PodGroupTemplateReference), scope)
	}); err != nil {
		return err
	}
	if err := s.AddGeneratedConversionFunc((*scheduling.PodGroupTemplateReference)(nil), (*schedulingv1alpha2.PodGroupTemplateReference)(nil), func(a, b interface{}, scope conversion.Scope) error {
		return Convert_scheduling_PodGroupTemplateReference_To_v1alpha2_PodGroupTemplateReference(a.(*scheduling.PodGroupTemplateReference), b.(*schedulingv1alpha2.PodGroupTemplateReference), scope)
	}); err != nil {
		return err
	}
	if err := s.AddGeneratedConversionFunc((*schedulingv1alpha2.TopologyConstraint)(nil), (*scheduling.TopologyConstraint)(nil), func(a, b interface{}, scope conversion.Scope) error {
		return Convert_v1alpha2_TopologyConstraint_To_scheduling_TopologyConstraint(a.(*schedulingv1alpha2.TopologyConstraint), b.(*scheduling.TopologyConstraint), scope)
	}); err != nil {
		return err
	}
	if err := s.AddGeneratedConversionFunc((*scheduling.TopologyConstraint)(nil), (*schedulingv1alpha2.TopologyConstraint)(nil), func(a, b interface{}, scope conversion.Scope) error {
		return Convert_scheduling_TopologyConstraint_To_v1alpha2_TopologyConstraint(a.(*scheduling.TopologyConstraint), b.(*schedulingv1alpha2.TopologyConstraint), scope)
	}); err != nil {
		return err
	}
	if err := s.AddGeneratedConversionFunc((*schedulingv1alpha2.TypedLocalObjectReference)(nil), (*scheduling.TypedLocalObjectReference)(nil), func(a, b interface{}, scope conversion.Scope) error {
		return Convert_v1alpha2_TypedLocalObjectReference_To_scheduling_TypedLocalObjectReference(a.(*schedulingv1alpha2.TypedLocalObjectReference), b.(*scheduling.TypedLocalObjectReference), scope)
	}); err != nil {
		return err
	}
	if err := s.AddGeneratedConversionFunc((*scheduling.TypedLocalObjectReference)(nil), (*schedulingv1alpha2.TypedLocalObjectReference)(nil), func(a, b interface{}, scope conversion.Scope) error {
		return Convert_scheduling_TypedLocalObjectReference_To_v1alpha2_TypedLocalObjectReference(a.(*scheduling.TypedLocalObjectReference), b.(*schedulingv1alpha2.TypedLocalObjectReference), scope)
	}); err != nil {
		return err
	}
	if err := s.AddGeneratedConversionFunc((*schedulingv1alpha2.Workload)(nil), (*scheduling.Workload)(nil), func(a, b interface{}, scope conversion.Scope) error {
		return Convert_v1alpha2_Workload_To_scheduling_Workload(a.(*schedulingv1alpha2.Workload), b.(*scheduling.Workload), scope)
	}); err != nil {
		return err
	}
	if err := s.AddGeneratedConversionFunc((*scheduling.Workload)(nil), (*schedulingv1alpha2.Workload)(nil), func(a, b interface{}, scope conversion.Scope) error {
		return Convert_scheduling_Workload_To_v1alpha2_Workload(a.(*scheduling.Workload), b.(*schedulingv1alpha2.Workload), scope)
	}); err != nil {
		return err
	}
	if err := s.AddGeneratedConversionFunc((*schedulingv1alpha2.WorkloadList)(nil), (*scheduling.WorkloadList)(nil), func(a, b interface{}, scope conversion.Scope) error {
		return Convert_v1alpha2_WorkloadList_To_scheduling_WorkloadList(a.(*schedulingv1alpha2.WorkloadList), b.(*scheduling.WorkloadList), scope)
	}); err != nil {
		return err
	}
	if err := s.AddGeneratedConversionFunc((*scheduling.WorkloadList)(nil), (*schedulingv1alpha2.WorkloadList)(nil), func(a, b interface{}, scope conversion.Scope) error {
		return Convert_scheduling_WorkloadList_To_v1alpha2_WorkloadList(a.(*scheduling.WorkloadList), b.(*schedulingv1alpha2.WorkloadList), scope)
	}); err != nil {
		return err
	}
	if err := s.AddGeneratedConversionFunc((*schedulingv1alpha2.WorkloadPodGroupTemplateReference)(nil), (*scheduling.WorkloadPodGroupTemplateReference)(nil), func(a, b interface{}, scope conversion.Scope) error {
		return Convert_v1alpha2_WorkloadPodGroupTemplateReference_To_scheduling_WorkloadPodGroupTemplateReference(a.(*schedulingv1alpha2.WorkloadPodGroupTemplateReference), b.(*scheduling.WorkloadPodGroupTemplateReference), scope)
	}); err != nil {
		return err
	}
	if err := s.AddGeneratedConversionFunc((*scheduling.WorkloadPodGroupTemplateReference)(nil), (*schedulingv1alpha2.WorkloadPodGroupTemplateReference)(nil), func(a, b interface{}, scope conversion.Scope) error {
		return Convert_scheduling_WorkloadPodGroupTemplateReference_To_v1alpha2_WorkloadPodGroupTemplateReference(a.(*scheduling.WorkloadPodGroupTemplateReference), b.(*schedulingv1alpha2.WorkloadPodGroupTemplateReference), scope)
	}); err != nil {
		return err
	}
	if err := s.AddGeneratedConversionFunc((*schedulingv1alpha2.WorkloadSpec)(nil), (*scheduling.WorkloadSpec)(nil), func(a, b interface{}, scope conversion.Scope) error {
		return Convert_v1alpha2_WorkloadSpec_To_scheduling_WorkloadSpec(a.(*schedulingv1alpha2.WorkloadSpec), b.(*scheduling.WorkloadSpec), scope)
	}); err != nil {
		return err
	}
	if err := s.AddGeneratedConversionFunc((*scheduling.WorkloadSpec)(nil), (*schedulingv1alpha2.WorkloadSpec)(nil), func(a, b interface{}, scope conversion.Scope) error {
		return Convert_scheduling_WorkloadSpec_To_v1alpha2_WorkloadSpec(a.(*scheduling.WorkloadSpec), b.(*schedulingv1alpha2.WorkloadSpec), scope)
	}); err != nil {
		return err
	}
	return nil
}

func autoConvert_v1alpha2_BasicSchedulingPolicy_To_scheduling_BasicSchedulingPolicy(in *schedulingv1alpha2.BasicSchedulingPolicy, out *scheduling.BasicSchedulingPolicy, s conversion.Scope) error {
	return nil
}

// Convert_v1alpha2_BasicSchedulingPolicy_To_scheduling_BasicSchedulingPolicy is an autogenerated conversion function.
func Convert_v1alpha2_BasicSchedulingPolicy_To_scheduling_BasicSchedulingPolicy(in *schedulingv1alpha2.BasicSchedulingPolicy, out *scheduling.BasicSchedulingPolicy, s conversion.Scope) error {
	return autoConvert_v1alpha2_BasicSchedulingPolicy_To_scheduling_BasicSchedulingPolicy(in, out, s)
}

func autoConvert_scheduling_BasicSchedulingPolicy_To_v1alpha2_BasicSchedulingPolicy(in *scheduling.BasicSchedulingPolicy, out *schedulingv1alpha2.BasicSchedulingPolicy, s conversion.Scope) error {
	return nil
}

// Convert_scheduling_BasicSchedulingPolicy_To_v1alpha2_BasicSchedulingPolicy is an autogenerated conversion function.
func Convert_scheduling_BasicSchedulingPolicy_To_v1alpha2_BasicSchedulingPolicy(in *scheduling.BasicSchedulingPolicy, out *schedulingv1alpha2.BasicSchedulingPolicy, s conversion.Scope) error {
	return autoConvert_scheduling_BasicSchedulingPolicy_To_v1alpha2_BasicSchedulingPolicy(in, out, s)
}

func autoConvert_v1alpha2_GangSchedulingPolicy_To_scheduling_GangSchedulingPolicy(in *schedulingv1alpha2.GangSchedulingPolicy, out *scheduling.GangSchedulingPolicy, s conversion.Scope) error {
	out.MinCount = in.MinCount
	return nil
}

// Convert_v1alpha2_GangSchedulingPolicy_To_scheduling_GangSchedulingPolicy is an autogenerated conversion function.
func Convert_v1alpha2_GangSchedulingPolicy_To_scheduling_GangSchedulingPolicy(in *schedulingv1alpha2.GangSchedulingPolicy, out *scheduling.GangSchedulingPolicy, s conversion.Scope) error {
	return autoConvert_v1alpha2_GangSchedulingPolicy_To_scheduling_GangSchedulingPolicy(in, out, s)
}

func autoConvert_scheduling_GangSchedulingPolicy_To_v1alpha2_GangSchedulingPolicy(in *scheduling.GangSchedulingPolicy, out *schedulingv1alpha2.GangSchedulingPolicy, s conversion.Scope) error {
	out.MinCount = in.MinCount
	return nil
}

// Convert_scheduling_GangSchedulingPolicy_To_v1alpha2_GangSchedulingPolicy is an autogenerated conversion function.
func Convert_scheduling_GangSchedulingPolicy_To_v1alpha2_GangSchedulingPolicy(in *scheduling.GangSchedulingPolicy, out *schedulingv1alpha2.GangSchedulingPolicy, s conversion.Scope) error {
	return autoConvert_scheduling_GangSchedulingPolicy_To_v1alpha2_GangSchedulingPolicy(in, out, s)
}

func autoConvert_v1alpha2_PodGroup_To_scheduling_PodGroup(in *schedulingv1alpha2.PodGroup, out *scheduling.PodGroup, s conversion.Scope) error {
	out.ObjectMeta = in.ObjectMeta
	if err := Convert_v1alpha2_PodGroupSpec_To_scheduling_PodGroupSpec(&in.Spec, &out.Spec, s); err != nil {
		return err
	}
	if err := Convert_v1alpha2_PodGroupStatus_To_scheduling_PodGroupStatus(&in.Status, &out.Status, s); err != nil {
		return err
	}
	return nil
}

// Convert_v1alpha2_PodGroup_To_scheduling_PodGroup is an autogenerated conversion function.
func Convert_v1alpha2_PodGroup_To_scheduling_PodGroup(in *schedulingv1alpha2.PodGroup, out *scheduling.PodGroup, s conversion.Scope) error {
	return autoConvert_v1alpha2_PodGroup_To_scheduling_PodGroup(in, out, s)
}

func autoConvert_scheduling_PodGroup_To_v1alpha2_PodGroup(in *scheduling.PodGroup, out *schedulingv1alpha2.PodGroup, s conversion.Scope) error {
	out.ObjectMeta = in.ObjectMeta
	if err := Convert_scheduling_PodGroupSpec_To_v1alpha2_PodGroupSpec(&in.Spec, &out.Spec, s); err != nil {
		return err
	}
	if err := Convert_scheduling_PodGroupStatus_To_v1alpha2_PodGroupStatus(&in.Status, &out.Status, s); err != nil {
		return err
	}
	return nil
}

// Convert_scheduling_PodGroup_To_v1alpha2_PodGroup is an autogenerated conversion function.
func Convert_scheduling_PodGroup_To_v1alpha2_PodGroup(in *scheduling.PodGroup, out *schedulingv1alpha2.PodGroup, s conversion.Scope) error {
	return autoConvert_scheduling_PodGroup_To_v1alpha2_PodGroup(in, out, s)
}

func autoConvert_v1alpha2_PodGroupList_To_scheduling_PodGroupList(in *schedulingv1alpha2.PodGroupList, out *scheduling.PodGroupList, s conversion.Scope) error {
	out.ListMeta = in.ListMeta
	out.Items = *(*[]scheduling.PodGroup)(unsafe.Pointer(&in.Items))
	return nil
}

// Convert_v1alpha2_PodGroupList_To_scheduling_PodGroupList is an autogenerated conversion function.
func Convert_v1alpha2_PodGroupList_To_scheduling_PodGroupList(in *schedulingv1alpha2.PodGroupList, out *scheduling.PodGroupList, s conversion.Scope) error {
	return autoConvert_v1alpha2_PodGroupList_To_scheduling_PodGroupList(in, out, s)
}

func autoConvert_scheduling_PodGroupList_To_v1alpha2_PodGroupList(in *scheduling.PodGroupList, out *schedulingv1alpha2.PodGroupList, s conversion.Scope) error {
	out.ListMeta = in.ListMeta
	out.Items = *(*[]schedulingv1alpha2.PodGroup)(unsafe.Pointer(&in.Items))
	return nil
}

// Convert_scheduling_PodGroupList_To_v1alpha2_PodGroupList is an autogenerated conversion function.
func Convert_scheduling_PodGroupList_To_v1alpha2_PodGroupList(in *scheduling.PodGroupList, out *schedulingv1alpha2.PodGroupList, s conversion.Scope) error {
	return autoConvert_scheduling_PodGroupList_To_v1alpha2_PodGroupList(in, out, s)
}

func autoConvert_v1alpha2_PodGroupResourceClaim_To_scheduling_PodGroupResourceClaim(in *schedulingv1alpha2.PodGroupResourceClaim, out *scheduling.PodGroupResourceClaim, s conversion.Scope) error {
	out.Name = in.Name
	out.ResourceClaimName = (*string)(unsafe.Pointer(in.ResourceClaimName))
	out.ResourceClaimTemplateName = (*string)(unsafe.Pointer(in.ResourceClaimTemplateName))
	return nil
}

// Convert_v1alpha2_PodGroupResourceClaim_To_scheduling_PodGroupResourceClaim is an autogenerated conversion function.
func Convert_v1alpha2_PodGroupResourceClaim_To_scheduling_PodGroupResourceClaim(in *schedulingv1alpha2.PodGroupResourceClaim, out *scheduling.PodGroupResourceClaim, s conversion.Scope) error {
	return autoConvert_v1alpha2_PodGroupResourceClaim_To_scheduling_PodGroupResourceClaim(in, out, s)
}

func autoConvert_scheduling_PodGroupResourceClaim_To_v1alpha2_PodGroupResourceClaim(in *scheduling.PodGroupResourceClaim, out *schedulingv1alpha2.PodGroupResourceClaim, s conversion.Scope) error {
	out.Name = in.Name
	out.ResourceClaimName = (*string)(unsafe.Pointer(in.ResourceClaimName))
	out.ResourceClaimTemplateName = (*string)(unsafe.Pointer(in.ResourceClaimTemplateName))
	return nil
}

// Convert_scheduling_PodGroupResourceClaim_To_v1alpha2_PodGroupResourceClaim is an autogenerated conversion function.
func Convert_scheduling_PodGroupResourceClaim_To_v1alpha2_PodGroupResourceClaim(in *scheduling.PodGroupResourceClaim, out *schedulingv1alpha2.PodGroupResourceClaim, s conversion.Scope) error {
	return autoConvert_scheduling_PodGroupResourceClaim_To_v1alpha2_PodGroupResourceClaim(in, out, s)
}

func autoConvert_v1alpha2_PodGroupResourceClaimStatus_To_scheduling_PodGroupResourceClaimStatus(in *schedulingv1alpha2.PodGroupResourceClaimStatus, out *scheduling.PodGroupResourceClaimStatus, s conversion.Scope) error {
	out.Name = in.Name
	out.ResourceClaimName = (*string)(unsafe.Pointer(in.ResourceClaimName))
	return nil
}

// Convert_v1alpha2_PodGroupResourceClaimStatus_To_scheduling_PodGroupResourceClaimStatus is an autogenerated conversion function.
func Convert_v1alpha2_PodGroupResourceClaimStatus_To_scheduling_PodGroupResourceClaimStatus(in *schedulingv1alpha2.PodGroupResourceClaimStatus, out *scheduling.PodGroupResourceClaimStatus, s conversion.Scope) error {
	return autoConvert_v1alpha2_PodGroupResourceClaimStatus_To_scheduling_PodGroupResourceClaimStatus(in, out, s)
}

func autoConvert_scheduling_PodGroupResourceClaimStatus_To_v1alpha2_PodGroupResourceClaimStatus(in *scheduling.PodGroupResourceClaimStatus, out *schedulingv1alpha2.PodGroupResourceClaimStatus, s conversion.Scope) error {
	out.Name = in.Name
	out.ResourceClaimName = (*string)(unsafe.Pointer(in.ResourceClaimName))
	return nil
}

// Convert_scheduling_PodGroupResourceClaimStatus_To_v1alpha2_PodGroupResourceClaimStatus is an autogenerated conversion function.
func Convert_scheduling_PodGroupResourceClaimStatus_To_v1alpha2_PodGroupResourceClaimStatus(in *scheduling.PodGroupResourceClaimStatus, out *schedulingv1alpha2.PodGroupResourceClaimStatus, s conversion.Scope) error {
	return autoConvert_scheduling_PodGroupResourceClaimStatus_To_v1alpha2_PodGroupResourceClaimStatus(in, out, s)
}

func autoConvert_v1alpha2_PodGroupSchedulingConstraints_To_scheduling_PodGroupSchedulingConstraints(in *schedulingv1alpha2.PodGroupSchedulingConstraints, out *scheduling.PodGroupSchedulingConstraints, s conversion.Scope) error {
	out.Topology = *(*[]scheduling.TopologyConstraint)(unsafe.Pointer(&in.Topology))
	return nil
}

// Convert_v1alpha2_PodGroupSchedulingConstraints_To_scheduling_PodGroupSchedulingConstraints is an autogenerated conversion function.
func Convert_v1alpha2_PodGroupSchedulingConstraints_To_scheduling_PodGroupSchedulingConstraints(in *schedulingv1alpha2.PodGroupSchedulingConstraints, out *scheduling.PodGroupSchedulingConstraints, s conversion.Scope) error {
	return autoConvert_v1alpha2_PodGroupSchedulingConstraints_To_scheduling_PodGroupSchedulingConstraints(in, out, s)
}

func autoConvert_scheduling_PodGroupSchedulingConstraints_To_v1alpha2_PodGroupSchedulingConstraints(in *scheduling.PodGroupSchedulingConstraints, out *schedulingv1alpha2.PodGroupSchedulingConstraints, s conversion.Scope) error {
	out.Topology = *(*[]schedulingv1alpha2.TopologyConstraint)(unsafe.Pointer(&in.Topology))
	return nil
}

// Convert_scheduling_PodGroupSchedulingConstraints_To_v1alpha2_PodGroupSchedulingConstraints is an autogenerated conversion function.
func Convert_scheduling_PodGroupSchedulingConstraints_To_v1alpha2_PodGroupSchedulingConstraints(in *scheduling.PodGroupSchedulingConstraints, out *schedulingv1alpha2.PodGroupSchedulingConstraints, s conversion.Scope) error {
	return autoConvert_scheduling_PodGroupSchedulingConstraints_To_v1alpha2_PodGroupSchedulingConstraints(in, out, s)
}

func autoConvert_v1alpha2_PodGroupSchedulingPolicy_To_scheduling_PodGroupSchedulingPolicy(in *schedulingv1alpha2.PodGroupSchedulingPolicy, out *scheduling.PodGroupSchedulingPolicy, s conversion.Scope) error {
	out.Basic = (*scheduling.BasicSchedulingPolicy)(unsafe.Pointer(in.Basic))
	out.Gang = (*scheduling.GangSchedulingPolicy)(unsafe.Pointer(in.Gang))
	return nil
}

// Convert_v1alpha2_PodGroupSchedulingPolicy_To_scheduling_PodGroupSchedulingPolicy is an autogenerated conversion function.
func Convert_v1alpha2_PodGroupSchedulingPolicy_To_scheduling_PodGroupSchedulingPolicy(in *schedulingv1alpha2.PodGroupSchedulingPolicy, out *scheduling.PodGroupSchedulingPolicy, s conversion.Scope) error {
	return autoConvert_v1alpha2_PodGroupSchedulingPolicy_To_scheduling_PodGroupSchedulingPolicy(in, out, s)
}

func autoConvert_scheduling_PodGroupSchedulingPolicy_To_v1alpha2_PodGroupSchedulingPolicy(in *scheduling.PodGroupSchedulingPolicy, out *schedulingv1alpha2.PodGroupSchedulingPolicy, s conversion.Scope) error {
	out.Basic = (*schedulingv1alpha2.BasicSchedulingPolicy)(unsafe.Pointer(in.Basic))
	out.Gang = (*schedulingv1alpha2.GangSchedulingPolicy)(unsafe.Pointer(in.Gang))
	return nil
}

// Convert_scheduling_PodGroupSchedulingPolicy_To_v1alpha2_PodGroupSchedulingPolicy is an autogenerated conversion function.
func Convert_scheduling_PodGroupSchedulingPolicy_To_v1alpha2_PodGroupSchedulingPolicy(in *scheduling.PodGroupSchedulingPolicy, out *schedulingv1alpha2.PodGroupSchedulingPolicy, s conversion.Scope) error {
	return autoConvert_scheduling_PodGroupSchedulingPolicy_To_v1alpha2_PodGroupSchedulingPolicy(in, out, s)
}

func autoConvert_v1alpha2_PodGroupSpec_To_scheduling_PodGroupSpec(in *schedulingv1alpha2.PodGroupSpec, out *scheduling.PodGroupSpec, s conversion.Scope) error {
	out.PodGroupTemplateRef = (*scheduling.PodGroupTemplateReference)(unsafe.Pointer(in.PodGroupTemplateRef))
	if err := Convert_v1alpha2_PodGroupSchedulingPolicy_To_scheduling_PodGroupSchedulingPolicy(&in.SchedulingPolicy, &out.SchedulingPolicy, s); err != nil {
		return err
	}
	out.SchedulingConstraints = (*scheduling.PodGroupSchedulingConstraints)(unsafe.Pointer(in.SchedulingConstraints))
	out.ResourceClaims = *(*[]scheduling.PodGroupResourceClaim)(unsafe.Pointer(&in.ResourceClaims))
	out.DisruptionMode = (*scheduling.DisruptionMode)(unsafe.Pointer(in.DisruptionMode))
	out.PriorityClassName = in.PriorityClassName
	out.Priority = (*int32)(unsafe.Pointer(in.Priority))
	return nil
}

// Convert_v1alpha2_PodGroupSpec_To_scheduling_PodGroupSpec is an autogenerated conversion function.
func Convert_v1alpha2_PodGroupSpec_To_scheduling_PodGroupSpec(in *schedulingv1alpha2.PodGroupSpec, out *scheduling.PodGroupSpec, s conversion.Scope) error {
	return autoConvert_v1alpha2_PodGroupSpec_To_scheduling_PodGroupSpec(in, out, s)
}

func autoConvert_scheduling_PodGroupSpec_To_v1alpha2_PodGroupSpec(in *scheduling.PodGroupSpec, out *schedulingv1alpha2.PodGroupSpec, s conversion.Scope) error {
	out.PodGroupTemplateRef = (*schedulingv1alpha2.PodGroupTemplateReference)(unsafe.Pointer(in.PodGroupTemplateRef))
	if err := Convert_scheduling_PodGroupSchedulingPolicy_To_v1alpha2_PodGroupSchedulingPolicy(&in.SchedulingPolicy, &out.SchedulingPolicy, s); err != nil {
		return err
	}
	out.SchedulingConstraints = (*schedulingv1alpha2.PodGroupSchedulingConstraints)(unsafe.Pointer(in.SchedulingConstraints))
	out.ResourceClaims = *(*[]schedulingv1alpha2.PodGroupResourceClaim)(unsafe.Pointer(&in.ResourceClaims))
	out.DisruptionMode = (*schedulingv1alpha2.DisruptionMode)(unsafe.Pointer(in.DisruptionMode))
	out.PriorityClassName = in.PriorityClassName
	out.Priority = (*int32)(unsafe.Pointer(in.Priority))
	return nil
}

// Convert_scheduling_PodGroupSpec_To_v1alpha2_PodGroupSpec is an autogenerated conversion function.
func Convert_scheduling_PodGroupSpec_To_v1alpha2_PodGroupSpec(in *scheduling.PodGroupSpec, out *schedulingv1alpha2.PodGroupSpec, s conversion.Scope) error {
	return autoConvert_scheduling_PodGroupSpec_To_v1alpha2_PodGroupSpec(in, out, s)
}

func autoConvert_v1alpha2_PodGroupStatus_To_scheduling_PodGroupStatus(in *schedulingv1alpha2.PodGroupStatus, out *scheduling.PodGroupStatus, s conversion.Scope) error {
	out.Conditions = *(*[]v1.Condition)(unsafe.Pointer(&in.Conditions))
	out.ResourceClaimStatuses = *(*[]scheduling.PodGroupResourceClaimStatus)(unsafe.Pointer(&in.ResourceClaimStatuses))
	return nil
}

// Convert_v1alpha2_PodGroupStatus_To_scheduling_PodGroupStatus is an autogenerated conversion function.
func Convert_v1alpha2_PodGroupStatus_To_scheduling_PodGroupStatus(in *schedulingv1alpha2.PodGroupStatus, out *scheduling.PodGroupStatus, s conversion.Scope) error {
	return autoConvert_v1alpha2_PodGroupStatus_To_scheduling_PodGroupStatus(in, out, s)
}

func autoConvert_scheduling_PodGroupStatus_To_v1alpha2_PodGroupStatus(in *scheduling.PodGroupStatus, out *schedulingv1alpha2.PodGroupStatus, s conversion.Scope) error {
	out.Conditions = *(*[]v1.Condition)(unsafe.Pointer(&in.Conditions))
	out.ResourceClaimStatuses = *(*[]schedulingv1alpha2.PodGroupResourceClaimStatus)(unsafe.Pointer(&in.ResourceClaimStatuses))
	return nil
}

// Convert_scheduling_PodGroupStatus_To_v1alpha2_PodGroupStatus is an autogenerated conversion function.
func Convert_scheduling_PodGroupStatus_To_v1alpha2_PodGroupStatus(in *scheduling.PodGroupStatus, out *schedulingv1alpha2.PodGroupStatus, s conversion.Scope) error {
	return autoConvert_scheduling_PodGroupStatus_To_v1alpha2_PodGroupStatus(in, out, s)
}

func autoConvert_v1alpha2_PodGroupTemplate_To_scheduling_PodGroupTemplate(in *schedulingv1alpha2.PodGroupTemplate, out *scheduling.PodGroupTemplate, s conversion.Scope) error {
	out.Name = in.Name
	if err := Convert_v1alpha2_PodGroupSchedulingPolicy_To_scheduling_PodGroupSchedulingPolicy(&in.SchedulingPolicy, &out.SchedulingPolicy, s); err != nil {
		return err
	}
	out.SchedulingConstraints = (*scheduling.PodGroupSchedulingConstraints)(unsafe.Pointer(in.SchedulingConstraints))
	out.ResourceClaims = *(*[]scheduling.PodGroupResourceClaim)(unsafe.Pointer(&in.ResourceClaims))
	out.DisruptionMode = (*scheduling.DisruptionMode)(unsafe.Pointer(in.DisruptionMode))
	out.PriorityClassName = in.PriorityClassName
	out.Priority = (*int32)(unsafe.Pointer(in.Priority))
	return nil
}

// Convert_v1alpha2_PodGroupTemplate_To_scheduling_PodGroupTemplate is an autogenerated conversion function.
func Convert_v1alpha2_PodGroupTemplate_To_scheduling_PodGroupTemplate(in *schedulingv1alpha2.PodGroupTemplate, out *scheduling.PodGroupTemplate, s conversion.Scope) error {
	return autoConvert_v1alpha2_PodGroupTemplate_To_scheduling_PodGroupTemplate(in, out, s)
}

func autoConvert_scheduling_PodGroupTemplate_To_v1alpha2_PodGroupTemplate(in *scheduling.PodGroupTemplate, out *schedulingv1alpha2.PodGroupTemplate, s conversion.Scope) error {
	out.Name = in.Name
	if err := Convert_scheduling_PodGroupSchedulingPolicy_To_v1alpha2_PodGroupSchedulingPolicy(&in.SchedulingPolicy, &out.SchedulingPolicy, s); err != nil {
		return err
	}
	out.SchedulingConstraints = (*schedulingv1alpha2.PodGroupSchedulingConstraints)(unsafe.Pointer(in.SchedulingConstraints))
	out.ResourceClaims = *(*[]schedulingv1alpha2.PodGroupResourceClaim)(unsafe.Pointer(&in.ResourceClaims))
	out.DisruptionMode = (*schedulingv1alpha2.DisruptionMode)(unsafe.Pointer(in.DisruptionMode))
	out.PriorityClassName = in.PriorityClassName
	out.Priority = (*int32)(unsafe.Pointer(in.Priority))
	return nil
}

// Convert_scheduling_PodGroupTemplate_To_v1alpha2_PodGroupTemplate is an autogenerated conversion function.
func Convert_scheduling_PodGroupTemplate_To_v1alpha2_PodGroupTemplate(in *scheduling.PodGroupTemplate, out *schedulingv1alpha2.PodGroupTemplate, s conversion.Scope) error {
	return autoConvert_scheduling_PodGroupTemplate_To_v1alpha2_PodGroupTemplate(in, out, s)
}

func autoConvert_v1alpha2_PodGroupTemplateReference_To_scheduling_PodGroupTemplateReference(in *schedulingv1alpha2.PodGroupTemplateReference, out *scheduling.PodGroupTemplateReference, s conversion.Scope) error {
	out.Workload = (*scheduling.WorkloadPodGroupTemplateReference)(unsafe.Pointer(in.Workload))
	return nil
}

// Convert_v1alpha2_PodGroupTemplateReference_To_scheduling_PodGroupTemplateReference is an autogenerated conversion function.
func Convert_v1alpha2_PodGroupTemplateReference_To_scheduling_PodGroupTemplateReference(in *schedulingv1alpha2.PodGroupTemplateReference, out *scheduling.PodGroupTemplateReference, s conversion.Scope) error {
	return autoConvert_v1alpha2_PodGroupTemplateReference_To_scheduling_PodGroupTemplateReference(in, out, s)
}

func autoConvert_scheduling_PodGroupTemplateReference_To_v1alpha2_PodGroupTemplateReference(in *scheduling.PodGroupTemplateReference, out *schedulingv1alpha2.PodGroupTemplateReference, s conversion.Scope) error {
	out.Workload = (*schedulingv1alpha2.WorkloadPodGroupTemplateReference)(unsafe.Pointer(in.Workload))
	return nil
}

// Convert_scheduling_PodGroupTemplateReference_To_v1alpha2_PodGroupTemplateReference is an autogenerated conversion function.
func Convert_scheduling_PodGroupTemplateReference_To_v1alpha2_PodGroupTemplateReference(in *scheduling.PodGroupTemplateReference, out *schedulingv1alpha2.PodGroupTemplateReference, s conversion.Scope) error {
	return autoConvert_scheduling_PodGroupTemplateReference_To_v1alpha2_PodGroupTemplateReference(in, out, s)
}

func autoConvert_v1alpha2_TopologyConstraint_To_scheduling_TopologyConstraint(in *schedulingv1alpha2.TopologyConstraint, out *scheduling.TopologyConstraint, s conversion.Scope) error {
	out.Key = in.Key
	return nil
}

// Convert_v1alpha2_TopologyConstraint_To_scheduling_TopologyConstraint is an autogenerated conversion function.
func Convert_v1alpha2_TopologyConstraint_To_scheduling_TopologyConstraint(in *schedulingv1alpha2.TopologyConstraint, out *scheduling.TopologyConstraint, s conversion.Scope) error {
	return autoConvert_v1alpha2_TopologyConstraint_To_scheduling_TopologyConstraint(in, out, s)
}

func autoConvert_scheduling_TopologyConstraint_To_v1alpha2_TopologyConstraint(in *scheduling.TopologyConstraint, out *schedulingv1alpha2.TopologyConstraint, s conversion.Scope) error {
	out.Key = in.Key
	return nil
}

// Convert_scheduling_TopologyConstraint_To_v1alpha2_TopologyConstraint is an autogenerated conversion function.
func Convert_scheduling_TopologyConstraint_To_v1alpha2_TopologyConstraint(in *scheduling.TopologyConstraint, out *schedulingv1alpha2.TopologyConstraint, s conversion.Scope) error {
	return autoConvert_scheduling_TopologyConstraint_To_v1alpha2_TopologyConstraint(in, out, s)
}

func autoConvert_v1alpha2_TypedLocalObjectReference_To_scheduling_TypedLocalObjectReference(in *schedulingv1alpha2.TypedLocalObjectReference, out *scheduling.TypedLocalObjectReference, s conversion.Scope) error {
	out.APIGroup = in.APIGroup
	out.Kind = in.Kind
	out.Name = in.Name
	return nil
}

// Convert_v1alpha2_TypedLocalObjectReference_To_scheduling_TypedLocalObjectReference is an autogenerated conversion function.
func Convert_v1alpha2_TypedLocalObjectReference_To_scheduling_TypedLocalObjectReference(in *schedulingv1alpha2.TypedLocalObjectReference, out *scheduling.TypedLocalObjectReference, s conversion.Scope) error {
	return autoConvert_v1alpha2_TypedLocalObjectReference_To_scheduling_TypedLocalObjectReference(in, out, s)
}

func autoConvert_scheduling_TypedLocalObjectReference_To_v1alpha2_TypedLocalObjectReference(in *scheduling.TypedLocalObjectReference, out *schedulingv1alpha2.TypedLocalObjectReference, s conversion.Scope) error {
	out.APIGroup = in.APIGroup
	out.Kind = in.Kind
	out.Name = in.Name
	return nil
}

// Convert_scheduling_TypedLocalObjectReference_To_v1alpha2_TypedLocalObjectReference is an autogenerated conversion function.
func Convert_scheduling_TypedLocalObjectReference_To_v1alpha2_TypedLocalObjectReference(in *scheduling.TypedLocalObjectReference, out *schedulingv1alpha2.TypedLocalObjectReference, s conversion.Scope) error {
	return autoConvert_scheduling_TypedLocalObjectReference_To_v1alpha2_TypedLocalObjectReference(in, out, s)
}

func autoConvert_v1alpha2_Workload_To_scheduling_Workload(in *schedulingv1alpha2.Workload, out *scheduling.Workload, s conversion.Scope) error {
	out.ObjectMeta = in.ObjectMeta
	if err := Convert_v1alpha2_WorkloadSpec_To_scheduling_WorkloadSpec(&in.Spec, &out.Spec, s); err != nil {
		return err
	}
	return nil
}

// Convert_v1alpha2_Workload_To_scheduling_Workload is an autogenerated conversion function.
func Convert_v1alpha2_Workload_To_scheduling_Workload(in *schedulingv1alpha2.Workload, out *scheduling.Workload, s conversion.Scope) error {
	return autoConvert_v1alpha2_Workload_To_scheduling_Workload(in, out, s)
}

func autoConvert_scheduling_Workload_To_v1alpha2_Workload(in *scheduling.Workload, out *schedulingv1alpha2.Workload, s conversion.Scope) error {
	out.ObjectMeta = in.ObjectMeta
	if err := Convert_scheduling_WorkloadSpec_To_v1alpha2_WorkloadSpec(&in.Spec, &out.Spec, s); err != nil {
		return err
	}
	return nil
}

// Convert_scheduling_Workload_To_v1alpha2_Workload is an autogenerated conversion function.
func Convert_scheduling_Workload_To_v1alpha2_Workload(in *scheduling.Workload, out *schedulingv1alpha2.Workload, s conversion.Scope) error {
	return autoConvert_scheduling_Workload_To_v1alpha2_Workload(in, out, s)
}

func autoConvert_v1alpha2_WorkloadList_To_scheduling_WorkloadList(in *schedulingv1alpha2.WorkloadList, out *scheduling.WorkloadList, s conversion.Scope) error {
	out.ListMeta = in.ListMeta
	out.Items = *(*[]scheduling.Workload)(unsafe.Pointer(&in.Items))
	return nil
}

// Convert_v1alpha2_WorkloadList_To_scheduling_WorkloadList is an autogenerated conversion function.
func Convert_v1alpha2_WorkloadList_To_scheduling_WorkloadList(in *schedulingv1alpha2.WorkloadList, out *scheduling.WorkloadList, s conversion.Scope) error {
	return autoConvert_v1alpha2_WorkloadList_To_scheduling_WorkloadList(in, out, s)
}

func autoConvert_scheduling_WorkloadList_To_v1alpha2_WorkloadList(in *scheduling.WorkloadList, out *schedulingv1alpha2.WorkloadList, s conversion.Scope) error {
	out.ListMeta = in.ListMeta
	out.Items = *(*[]schedulingv1alpha2.Workload)(unsafe.Pointer(&in.Items))
	return nil
}

// Convert_scheduling_WorkloadList_To_v1alpha2_WorkloadList is an autogenerated conversion function.
func Convert_scheduling_WorkloadList_To_v1alpha2_WorkloadList(in *scheduling.WorkloadList, out *schedulingv1alpha2.WorkloadList, s conversion.Scope) error {
	return autoConvert_scheduling_WorkloadList_To_v1alpha2_WorkloadList(in, out, s)
}

func autoConvert_v1alpha2_WorkloadPodGroupTemplateReference_To_scheduling_WorkloadPodGroupTemplateReference(in *schedulingv1alpha2.WorkloadPodGroupTemplateReference, out *scheduling.WorkloadPodGroupTemplateReference, s conversion.Scope) error {
	out.WorkloadName = in.WorkloadName
	out.PodGroupTemplateName = in.PodGroupTemplateName
	return nil
}

// Convert_v1alpha2_WorkloadPodGroupTemplateReference_To_scheduling_WorkloadPodGroupTemplateReference is an autogenerated conversion function.
func Convert_v1alpha2_WorkloadPodGroupTemplateReference_To_scheduling_WorkloadPodGroupTemplateReference(in *schedulingv1alpha2.WorkloadPodGroupTemplateReference, out *scheduling.WorkloadPodGroupTemplateReference, s conversion.Scope) error {
	return autoConvert_v1alpha2_WorkloadPodGroupTemplateReference_To_scheduling_WorkloadPodGroupTemplateReference(in, out, s)
}

func autoConvert_scheduling_WorkloadPodGroupTemplateReference_To_v1alpha2_WorkloadPodGroupTemplateReference(in *scheduling.WorkloadPodGroupTemplateReference, out *schedulingv1alpha2.WorkloadPodGroupTemplateReference, s conversion.Scope) error {
	out.WorkloadName = in.WorkloadName
	out.PodGroupTemplateName = in.PodGroupTemplateName
	return nil
}

// Convert_scheduling_WorkloadPodGroupTemplateReference_To_v1alpha2_WorkloadPodGroupTemplateReference is an autogenerated conversion function.
func Convert_scheduling_WorkloadPodGroupTemplateReference_To_v1alpha2_WorkloadPodGroupTemplateReference(in *scheduling.WorkloadPodGroupTemplateReference, out *schedulingv1alpha2.WorkloadPodGroupTemplateReference, s conversion.Scope) error {
	return autoConvert_scheduling_WorkloadPodGroupTemplateReference_To_v1alpha2_WorkloadPodGroupTemplateReference(in, out, s)
}

func autoConvert_v1alpha2_WorkloadSpec_To_scheduling_WorkloadSpec(in *schedulingv1alpha2.WorkloadSpec, out *scheduling.WorkloadSpec, s conversion.Scope) error {
	out.ControllerRef = (*scheduling.TypedLocalObjectReference)(unsafe.Pointer(in.ControllerRef))
	out.PodGroupTemplates = *(*[]scheduling.PodGroupTemplate)(unsafe.Pointer(&in.PodGroupTemplates))
	return nil
}

// Convert_v1alpha2_WorkloadSpec_To_scheduling_WorkloadSpec is an autogenerated conversion function.
func Convert_v1alpha2_WorkloadSpec_To_scheduling_WorkloadSpec(in *schedulingv1alpha2.WorkloadSpec, out *scheduling.WorkloadSpec, s conversion.Scope) error {
	return autoConvert_v1alpha2_WorkloadSpec_To_scheduling_WorkloadSpec(in, out, s)
}

func autoConvert_scheduling_WorkloadSpec_To_v1alpha2_WorkloadSpec(in *scheduling.WorkloadSpec, out *schedulingv1alpha2.WorkloadSpec, s conversion.Scope) error {
	out.ControllerRef = (*schedulingv1alpha2.TypedLocalObjectReference)(unsafe.Pointer(in.ControllerRef))
	out.PodGroupTemplates = *(*[]schedulingv1alpha2.PodGroupTemplate)(unsafe.Pointer(&in.PodGroupTemplates))
	return nil
}

// Convert_scheduling_WorkloadSpec_To_v1alpha2_WorkloadSpec is an autogenerated conversion function.
func Convert_scheduling_WorkloadSpec_To_v1alpha2_WorkloadSpec(in *scheduling.WorkloadSpec, out *schedulingv1alpha2.WorkloadSpec, s conversion.Scope) error {
	return autoConvert_scheduling_WorkloadSpec_To_v1alpha2_WorkloadSpec(in, out, s)
}
