/*
Copyright The Kubernetes Authors.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
*/

//
//Copyright 2018 The Kubernetes Authors.
//
//Licensed under the Apache License, Version 2.0 (the "License");
//you may not use this file except in compliance with the License.
//You may obtain a copy of the License at
//
//http://www.apache.org/licenses/LICENSE-2.0
//
//Unless required by applicable law or agreed to in writing, software
//distributed under the License is distributed on an "AS IS" BASIS,
//WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
//See the License for the specific language governing permissions and
//limitations under the License.

// Code generated by protoc-gen-go. DO NOT EDIT.
// versions:
// 	protoc-gen-go v1.36.4
// 	protoc        v4.23.4
// source: pkg/kubelet/pluginmanager/pluginwatcher/example_plugin_apis/v1beta2/api.proto

package v1beta2

import (
	protoreflect "google.golang.org/protobuf/reflect/protoreflect"
	protoimpl "google.golang.org/protobuf/runtime/protoimpl"
	reflect "reflect"
	sync "sync"
	unsafe "unsafe"
)

const (
	// Verify that this generated code is sufficiently up-to-date.
	_ = protoimpl.EnforceVersion(20 - protoimpl.MinVersion)
	// Verify that runtime/protoimpl is sufficiently up-to-date.
	_ = protoimpl.EnforceVersion(protoimpl.MaxVersion - 20)
)

// Renames a field from v1beta1 ExampleRequest.
type ExampleRequest struct {
	state         protoimpl.MessageState `protogen:"open.v1"`
	Request       string                 `protobuf:"bytes,1,opt,name=request,proto3" json:"request,omitempty"`
	V1Beta2Field  string                 `protobuf:"bytes,2,opt,name=v1beta2_field,json=v1beta2Field,proto3" json:"v1beta2_field,omitempty"`
	unknownFields protoimpl.UnknownFields
	sizeCache     protoimpl.SizeCache
}

func (x *ExampleRequest) Reset() {
	*x = ExampleRequest{}
	mi := &file_pkg_kubelet_pluginmanager_pluginwatcher_example_plugin_apis_v1beta2_api_proto_msgTypes[0]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *ExampleRequest) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*ExampleRequest) ProtoMessage() {}

func (x *ExampleRequest) ProtoReflect() protoreflect.Message {
	mi := &file_pkg_kubelet_pluginmanager_pluginwatcher_example_plugin_apis_v1beta2_api_proto_msgTypes[0]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use ExampleRequest.ProtoReflect.Descriptor instead.
func (*ExampleRequest) Descriptor() ([]byte, []int) {
	return file_pkg_kubelet_pluginmanager_pluginwatcher_example_plugin_apis_v1beta2_api_proto_rawDescGZIP(), []int{0}
}

func (x *ExampleRequest) GetRequest() string {
	if x != nil {
		return x.Request
	}
	return ""
}

func (x *ExampleRequest) GetV1Beta2Field() string {
	if x != nil {
		return x.V1Beta2Field
	}
	return ""
}

type ExampleResponse struct {
	state         protoimpl.MessageState `protogen:"open.v1"`
	Error         string                 `protobuf:"bytes,1,opt,name=error,proto3" json:"error,omitempty"`
	unknownFields protoimpl.UnknownFields
	sizeCache     protoimpl.SizeCache
}

func (x *ExampleResponse) Reset() {
	*x = ExampleResponse{}
	mi := &file_pkg_kubelet_pluginmanager_pluginwatcher_example_plugin_apis_v1beta2_api_proto_msgTypes[1]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *ExampleResponse) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*ExampleResponse) ProtoMessage() {}

func (x *ExampleResponse) ProtoReflect() protoreflect.Message {
	mi := &file_pkg_kubelet_pluginmanager_pluginwatcher_example_plugin_apis_v1beta2_api_proto_msgTypes[1]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use ExampleResponse.ProtoReflect.Descriptor instead.
func (*ExampleResponse) Descriptor() ([]byte, []int) {
	return file_pkg_kubelet_pluginmanager_pluginwatcher_example_plugin_apis_v1beta2_api_proto_rawDescGZIP(), []int{1}
}

func (x *ExampleResponse) GetError() string {
	if x != nil {
		return x.Error
	}
	return ""
}

var File_pkg_kubelet_pluginmanager_pluginwatcher_example_plugin_apis_v1beta2_api_proto protoreflect.FileDescriptor

var file_pkg_kubelet_pluginmanager_pluginwatcher_example_plugin_apis_v1beta2_api_proto_rawDesc = string([]byte{
	0x0a, 0x4d, 0x70, 0x6b, 0x67, 0x2f, 0x6b, 0x75, 0x62, 0x65, 0x6c, 0x65, 0x74, 0x2f, 0x70, 0x6c,
	0x75, 0x67, 0x69, 0x6e, 0x6d, 0x61, 0x6e, 0x61, 0x67, 0x65, 0x72, 0x2f, 0x70, 0x6c, 0x75, 0x67,
	0x69, 0x6e, 0x77, 0x61, 0x74, 0x63, 0x68, 0x65, 0x72, 0x2f, 0x65, 0x78, 0x61, 0x6d, 0x70, 0x6c,
	0x65, 0x5f, 0x70, 0x6c, 0x75, 0x67, 0x69, 0x6e, 0x5f, 0x61, 0x70, 0x69, 0x73, 0x2f, 0x76, 0x31,
	0x62, 0x65, 0x74, 0x61, 0x32, 0x2f, 0x61, 0x70, 0x69, 0x2e, 0x70, 0x72, 0x6f, 0x74, 0x6f, 0x12,
	0x07, 0x76, 0x31, 0x62, 0x65, 0x74, 0x61, 0x32, 0x22, 0x4f, 0x0a, 0x0e, 0x45, 0x78, 0x61, 0x6d,
	0x70, 0x6c, 0x65, 0x52, 0x65, 0x71, 0x75, 0x65, 0x73, 0x74, 0x12, 0x18, 0x0a, 0x07, 0x72, 0x65,
	0x71, 0x75, 0x65, 0x73, 0x74, 0x18, 0x01, 0x20, 0x01, 0x28, 0x09, 0x52, 0x07, 0x72, 0x65, 0x71,
	0x75, 0x65, 0x73, 0x74, 0x12, 0x23, 0x0a, 0x0d, 0x76, 0x31, 0x62, 0x65, 0x74, 0x61, 0x32, 0x5f,
	0x66, 0x69, 0x65, 0x6c, 0x64, 0x18, 0x02, 0x20, 0x01, 0x28, 0x09, 0x52, 0x0c, 0x76, 0x31, 0x62,
	0x65, 0x74, 0x61, 0x32, 0x46, 0x69, 0x65, 0x6c, 0x64, 0x22, 0x27, 0x0a, 0x0f, 0x45, 0x78, 0x61,
	0x6d, 0x70, 0x6c, 0x65, 0x52, 0x65, 0x73, 0x70, 0x6f, 0x6e, 0x73, 0x65, 0x12, 0x14, 0x0a, 0x05,
	0x65, 0x72, 0x72, 0x6f, 0x72, 0x18, 0x01, 0x20, 0x01, 0x28, 0x09, 0x52, 0x05, 0x65, 0x72, 0x72,
	0x6f, 0x72, 0x32, 0x50, 0x0a, 0x07, 0x45, 0x78, 0x61, 0x6d, 0x70, 0x6c, 0x65, 0x12, 0x45, 0x0a,
	0x0e, 0x47, 0x65, 0x74, 0x45, 0x78, 0x61, 0x6d, 0x70, 0x6c, 0x65, 0x49, 0x6e, 0x66, 0x6f, 0x12,
	0x17, 0x2e, 0x76, 0x31, 0x62, 0x65, 0x74, 0x61, 0x32, 0x2e, 0x45, 0x78, 0x61, 0x6d, 0x70, 0x6c,
	0x65, 0x52, 0x65, 0x71, 0x75, 0x65, 0x73, 0x74, 0x1a, 0x18, 0x2e, 0x76, 0x31, 0x62, 0x65, 0x74,
	0x61, 0x32, 0x2e, 0x45, 0x78, 0x61, 0x6d, 0x70, 0x6c, 0x65, 0x52, 0x65, 0x73, 0x70, 0x6f, 0x6e,
	0x73, 0x65, 0x22, 0x00, 0x42, 0x57, 0x5a, 0x55, 0x6b, 0x38, 0x73, 0x2e, 0x69, 0x6f, 0x2f, 0x6b,
	0x75, 0x62, 0x65, 0x72, 0x6e, 0x65, 0x74, 0x65, 0x73, 0x2f, 0x70, 0x6b, 0x67, 0x2f, 0x6b, 0x75,
	0x62, 0x65, 0x6c, 0x65, 0x74, 0x2f, 0x70, 0x6c, 0x75, 0x67, 0x69, 0x6e, 0x6d, 0x61, 0x6e, 0x61,
	0x67, 0x65, 0x72, 0x2f, 0x70, 0x6c, 0x75, 0x67, 0x69, 0x6e, 0x77, 0x61, 0x74, 0x63, 0x68, 0x65,
	0x72, 0x2f, 0x65, 0x78, 0x61, 0x6d, 0x70, 0x6c, 0x65, 0x5f, 0x70, 0x6c, 0x75, 0x67, 0x69, 0x6e,
	0x5f, 0x61, 0x70, 0x69, 0x73, 0x2f, 0x76, 0x31, 0x62, 0x65, 0x74, 0x61, 0x32, 0x62, 0x06, 0x70,
	0x72, 0x6f, 0x74, 0x6f, 0x33,
})

var (
	file_pkg_kubelet_pluginmanager_pluginwatcher_example_plugin_apis_v1beta2_api_proto_rawDescOnce sync.Once
	file_pkg_kubelet_pluginmanager_pluginwatcher_example_plugin_apis_v1beta2_api_proto_rawDescData []byte
)

func file_pkg_kubelet_pluginmanager_pluginwatcher_example_plugin_apis_v1beta2_api_proto_rawDescGZIP() []byte {
	file_pkg_kubelet_pluginmanager_pluginwatcher_example_plugin_apis_v1beta2_api_proto_rawDescOnce.Do(func() {
		file_pkg_kubelet_pluginmanager_pluginwatcher_example_plugin_apis_v1beta2_api_proto_rawDescData = protoimpl.X.CompressGZIP(unsafe.Slice(unsafe.StringData(file_pkg_kubelet_pluginmanager_pluginwatcher_example_plugin_apis_v1beta2_api_proto_rawDesc), len(file_pkg_kubelet_pluginmanager_pluginwatcher_example_plugin_apis_v1beta2_api_proto_rawDesc)))
	})
	return file_pkg_kubelet_pluginmanager_pluginwatcher_example_plugin_apis_v1beta2_api_proto_rawDescData
}

var file_pkg_kubelet_pluginmanager_pluginwatcher_example_plugin_apis_v1beta2_api_proto_msgTypes = make([]protoimpl.MessageInfo, 2)
var file_pkg_kubelet_pluginmanager_pluginwatcher_example_plugin_apis_v1beta2_api_proto_goTypes = []any{
	(*ExampleRequest)(nil),  // 0: v1beta2.ExampleRequest
	(*ExampleResponse)(nil), // 1: v1beta2.ExampleResponse
}
var file_pkg_kubelet_pluginmanager_pluginwatcher_example_plugin_apis_v1beta2_api_proto_depIdxs = []int32{
	0, // 0: v1beta2.Example.GetExampleInfo:input_type -> v1beta2.ExampleRequest
	1, // 1: v1beta2.Example.GetExampleInfo:output_type -> v1beta2.ExampleResponse
	1, // [1:2] is the sub-list for method output_type
	0, // [0:1] is the sub-list for method input_type
	0, // [0:0] is the sub-list for extension type_name
	0, // [0:0] is the sub-list for extension extendee
	0, // [0:0] is the sub-list for field type_name
}

func init() {
	file_pkg_kubelet_pluginmanager_pluginwatcher_example_plugin_apis_v1beta2_api_proto_init()
}
func file_pkg_kubelet_pluginmanager_pluginwatcher_example_plugin_apis_v1beta2_api_proto_init() {
	if File_pkg_kubelet_pluginmanager_pluginwatcher_example_plugin_apis_v1beta2_api_proto != nil {
		return
	}
	type x struct{}
	out := protoimpl.TypeBuilder{
		File: protoimpl.DescBuilder{
			GoPackagePath: reflect.TypeOf(x{}).PkgPath(),
			RawDescriptor: unsafe.Slice(unsafe.StringData(file_pkg_kubelet_pluginmanager_pluginwatcher_example_plugin_apis_v1beta2_api_proto_rawDesc), len(file_pkg_kubelet_pluginmanager_pluginwatcher_example_plugin_apis_v1beta2_api_proto_rawDesc)),
			NumEnums:      0,
			NumMessages:   2,
			NumExtensions: 0,
			NumServices:   1,
		},
		GoTypes:           file_pkg_kubelet_pluginmanager_pluginwatcher_example_plugin_apis_v1beta2_api_proto_goTypes,
		DependencyIndexes: file_pkg_kubelet_pluginmanager_pluginwatcher_example_plugin_apis_v1beta2_api_proto_depIdxs,
		MessageInfos:      file_pkg_kubelet_pluginmanager_pluginwatcher_example_plugin_apis_v1beta2_api_proto_msgTypes,
	}.Build()
	File_pkg_kubelet_pluginmanager_pluginwatcher_example_plugin_apis_v1beta2_api_proto = out.File
	file_pkg_kubelet_pluginmanager_pluginwatcher_example_plugin_apis_v1beta2_api_proto_goTypes = nil
	file_pkg_kubelet_pluginmanager_pluginwatcher_example_plugin_apis_v1beta2_api_proto_depIdxs = nil
}
