/*
Copyright The Kubernetes Authors.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
*/

// Code generated by applyconfiguration-gen. DO NOT EDIT.

package v1

// PodIPApplyConfiguration represents a declarative configuration of the PodIP type for use
// with apply.
//
// PodIP represents a single IP address allocated to the pod.
type PodIPApplyConfiguration struct {
	// IP is the IP address assigned to the pod
	IP *string `json:"ip,omitempty"`
}

// PodIPApplyConfiguration constructs a declarative configuration of the PodIP type for use with
// apply.
func PodIP() *PodIPApplyConfiguration {
	return &PodIPApplyConfiguration{}
}

// WithIP sets the IP field in the declarative configuration to the given value
// and returns the receiver, so that objects can be built by chaining "With" function invocations.
// If called multiple times, the IP field is set to the value of the last call.
func (b *PodIPApplyConfiguration) WithIP(value string) *PodIPApplyConfiguration {
	b.IP = &value
	return b
}
